"""Tests for utils."""

import collections
import copy
import datetime
import logging
import os
import shutil
import unittest
from unittest.mock import patch

import numpy as np
import oyaml as yaml

import aspecd.exceptions
import aspecd.utils
from aspecd import utils, dataset


class TestFullClassName(unittest.TestCase):
    def test_full_class_name(self):
        dataset_ = aspecd.dataset.Dataset()
        class_name = utils.full_class_name(dataset_)
        self.assertEqual(class_name, "aspecd.dataset.Dataset")


class TestObjectFromClassName(unittest.TestCase):
    def test_object_from_class_name(self):
        class_name = "aspecd.dataset.Dataset"
        object_ = utils.object_from_class_name(class_name)
        self.assertTrue(isinstance(object_, dataset.Dataset))


class TestToDictMixin(unittest.TestCase):
    def setUp(self):
        class MixedIn(utils.ToDictMixin):
            pass

        self.mixed_in = MixedIn()

    @staticmethod
    def set_properties_from_dict(obj=None, dict_=None):
        dict_ = copy.deepcopy(dict_)
        for key in dict_:
            setattr(obj, key, dict_[key])

    def test_instantiate_class(self):
        pass

    def test_string_property(self):
        orig_dict = {"foo": "bar"}
        self.set_properties_from_dict(obj=self.mixed_in, dict_=orig_dict)
        obj_dict = self.mixed_in.to_dict()
        self.assertDictEqual(orig_dict, obj_dict)

    def test_only_public_properties(self):
        orig_dict = {"foo": "bar"}
        self.set_properties_from_dict(obj=self.mixed_in, dict_=orig_dict)
        self.mixed_in._protected_property = None
        obj_dict = self.mixed_in.to_dict()
        self.assertDictEqual(orig_dict, obj_dict)

    def test_dict_property(self):
        orig_dict = {"foo": {"bar": "baz"}}
        self.set_properties_from_dict(obj=self.mixed_in, dict_=orig_dict)
        obj_dict = self.mixed_in.to_dict()
        self.assertDictEqual(orig_dict, obj_dict)

    def test_cascaded_dict_property(self):
        orig_dict = {"foo": {"bar": {"baz": "baf"}}}
        self.set_properties_from_dict(obj=self.mixed_in, dict_=orig_dict)
        obj_dict = self.mixed_in.to_dict()
        self.assertDictEqual(orig_dict, obj_dict)

    def test_ordered_dict_property(self):
        orig_dict = {"foo": collections.OrderedDict({"bar": "baz"})}
        self.set_properties_from_dict(obj=self.mixed_in, dict_=orig_dict)
        obj_dict = self.mixed_in.to_dict()
        self.assertDictEqual(orig_dict, obj_dict)

    def test_mixed_in_object_property(self):
        orig_dict = {"foo": "bar"}
        mixed_in = utils.ToDictMixin()
        self.set_properties_from_dict(obj=mixed_in, dict_=orig_dict)
        self.mixed_in.object = mixed_in
        full_dict = {"object": {"foo": "bar"}}
        obj_dict = self.mixed_in.to_dict()
        self.assertDictEqual(full_dict, obj_dict)

    def test_generic_object_property(self):
        orig_dict = {"foo": "bar"}

        class DummyClass:
            pass

        obj = DummyClass()
        self.set_properties_from_dict(obj=obj, dict_=orig_dict)
        self.mixed_in.object = obj
        full_dict = {"object": {"foo": "bar"}}
        obj_dict = self.mixed_in.to_dict()
        self.assertDictEqual(full_dict, obj_dict)

    def test_list_of_dicts_property(self):
        orig_dict = {"foo": [{"foo": "bar"}, {"bar": "baz"}]}
        self.set_properties_from_dict(obj=self.mixed_in, dict_=orig_dict)
        obj_dict = self.mixed_in.to_dict()
        self.assertDictEqual(orig_dict, obj_dict)

    def test_list_of_dicts_containing_dicts_property(self):
        orig_dict = {"foo": [{"foo": {"foobar": "bar"}}, {"bar": "baz"}]}
        self.set_properties_from_dict(obj=self.mixed_in, dict_=orig_dict)
        obj_dict = self.mixed_in.to_dict()
        self.assertDictEqual(orig_dict, obj_dict)

    def test_list_of_objects_property(self):
        toobj_dict = {"foo": "bar"}
        obj1 = utils.ToDictMixin()
        obj2 = utils.ToDictMixin()
        self.set_properties_from_dict(obj=obj1, dict_=toobj_dict)
        self.set_properties_from_dict(obj=obj2, dict_=toobj_dict)
        self.mixed_in.objects = [obj1, obj2]
        orig_dict = {"objects": [toobj_dict, toobj_dict]}
        obj_dict = self.mixed_in.to_dict()
        self.assertDictEqual(orig_dict, obj_dict)

    def test_list_of_dicts_containing_objects_property(self):
        toobj_dict = {"foo": "bar"}
        obj1 = utils.ToDictMixin()
        obj2 = utils.ToDictMixin()
        self.set_properties_from_dict(obj=obj1, dict_=toobj_dict)
        self.set_properties_from_dict(obj=obj2, dict_=toobj_dict)
        orig_dict = {"foo": [{"foo": toobj_dict}, {"bar": toobj_dict}]}
        self.set_properties_from_dict(obj=self.mixed_in, dict_=orig_dict)
        # noinspection PyUnresolvedReferences
        self.mixed_in.foo[0]["foo"] = obj1
        # noinspection PyUnresolvedReferences
        self.mixed_in.foo[1]["bar"] = obj2
        obj_dict = self.mixed_in.to_dict()
        self.assertEqual(orig_dict["foo"], obj_dict["foo"])

    def test_datetime_property(self):
        date = datetime.datetime.now()
        toobj_dict = {"date": date}
        self.set_properties_from_dict(obj=self.mixed_in, dict_=toobj_dict)
        orig_dict = {"date": str(date)}
        obj_dict = self.mixed_in.to_dict()
        self.assertDictEqual(orig_dict, obj_dict)

    def test_date_property(self):
        date = datetime.date.today()
        toobj_dict = {"date": date}
        self.set_properties_from_dict(obj=self.mixed_in, dict_=toobj_dict)
        orig_dict = {"date": str(date)}
        obj_dict = self.mixed_in.to_dict()
        self.assertDictEqual(orig_dict, obj_dict)

    def test_time_property(self):
        date = datetime.time(12, 10, 30)
        toobj_dict = {"date": date}
        self.set_properties_from_dict(obj=self.mixed_in, dict_=toobj_dict)
        orig_dict = {"date": str(date)}
        obj_dict = self.mixed_in.to_dict()
        self.assertDictEqual(orig_dict, obj_dict)

    def test_has_odict_attribute(self):
        self.assertTrue(hasattr(self.mixed_in, "__odict__"))

    def test_odict_attribute_is_ordered_dict(self):
        self.assertTrue(
            isinstance(self.mixed_in.__odict__, collections.OrderedDict)
        )

    def test_odict_preserves_argument_definition_order(self):
        arguments = ["purpose", "operator", "labbook"]

        class Test(utils.ToDictMixin):
            def __init__(self):
                super().__init__()
                self.purpose = ""
                self.operator = ""
                self.labbook = ""

        obj = Test()
        self.assertEqual(arguments, list(obj.to_dict().keys()))

    def test_with_properties_to_exclude(self):
        class Test(utils.ToDictMixin):
            def __init__(self):
                super().__init__()
                self.purpose = ""
                self.operator = ""
                self._exclude_from_to_dict = ["operator"]

        obj = Test()
        self.assertEqual(["purpose"], list(obj.to_dict().keys()))

    def test_with_property_to_include(self):
        class Test(utils.ToDictMixin):
            def __init__(self):
                super().__init__()
                self.purpose = ""
                self._foo = None
                self._include_in_to_dict = ["_foo"]

        obj = Test()
        self.assertEqual(["purpose", "_foo"], list(obj.to_dict().keys()))

    def test_with_properties_to_include(self):
        class Test(utils.ToDictMixin):
            def __init__(self):
                super().__init__()
                self.purpose = ""
                self._foo = None
                self._bar = None
                self._include_in_to_dict = ["_foo", "_bar"]

        obj = Test()
        self.assertEqual(
            ["purpose", "_foo", "_bar"], list(obj.to_dict().keys())
        )

    def test_remove_empty_properties_from_dict(self):
        class Test(utils.ToDictMixin):
            def __init__(self):
                super().__init__()
                self.purpose = ""
                self.comment = "Lorem ipsum"

        obj = Test()
        self.assertEqual(
            ["comment"], list(obj.to_dict(remove_empty=True).keys())
        )

    def test_properties_to_include_are_not_traversed(self):
        class Test(utils.ToDictMixin):
            def __init__(self):
                super().__init__()
                self.purpose = {}
                self._foo = None
                self._include_in_to_dict = ["_foo"]

        obj = Test()
        self.assertNotIn("_foo", list(obj.to_dict()["purpose"]))

    def test_properties_to_exclude_are_not_traversed(self):
        class Test(utils.ToDictMixin):
            def __init__(self):
                super().__init__()
                self.purpose = {"operator": "John Doe"}
                self.operator = ""
                self._exclude_from_to_dict = ["operator"]

        obj = Test()
        self.assertIn("operator", obj.to_dict()["purpose"])

    def test_properties_to_include_are_taken_from_object(self):
        class TestProperty(utils.ToDictMixin):
            def __init__(self):
                super().__init__()
                self.operator = "John Doe"
                self._bar = None
                self._include_in_to_dict = ["_bar"]

        class Test(utils.ToDictMixin):
            def __init__(self):
                super().__init__()
                self.purpose = TestProperty()
                self._foo = None
                self._include_in_to_dict = ["_foo"]

        obj = Test()
        self.assertIn("_bar", list(obj.to_dict()["purpose"]))

    def test_properties_to_exclude_are_taken_from_object(self):
        class TestProperty(utils.ToDictMixin):
            def __init__(self):
                super().__init__()
                self.operator = "John Doe"
                self._bar = None
                self._exclude_from_to_dict = ["_bar"]

        class Test(utils.ToDictMixin):
            def __init__(self):
                super().__init__()
                self.purpose = {"operator": "John Doe"}
                self.operator = TestProperty()
                self._exclude_from_to_dict = ["operator"]

        obj = Test()
        self.assertNotIn("_bar", obj.to_dict()["purpose"])

    def test_property_to_include_is_property(self):
        class Test(utils.ToDictMixin):
            def __init__(self):
                super().__init__()
                self.purpose = ""
                self._prop = "foo"
                self._include_in_to_dict = ["prop"]

            @property
            def prop(self):
                return self._prop

        obj = Test()
        self.assertEqual(["purpose", "prop"], list(obj.to_dict().keys()))


class TestGetAspecdVersion(unittest.TestCase):
    def test_version_not_empty(self):
        version = utils.get_aspecd_version()
        self.assertTrue(version)


class TestGetVersion(unittest.TestCase):
    def test_version_not_empty(self):
        version = utils.package_version("aspecd")
        self.assertTrue(version)

    def test_version_correct_for_aspecd_package(self):
        version = utils.package_version("aspecd")
        self.assertEqual(utils.get_aspecd_version(), version)


class TestPackageName(unittest.TestCase):
    def test_package_name_returns_correct_package(self):
        package_name = utils.package_name()
        self.assertEqual(package_name, "aspecd")

    def test_package_name_with_object_returns_correct_package(self):
        package_name = utils.package_name(unittest.TestCase())
        self.assertEqual(package_name, "unittest")


class TestConfigDir(unittest.TestCase):
    def test_config_dir_not_empty(self):
        config_dir = utils.config_dir()
        self.assertTrue(config_dir)


class TestYaml(unittest.TestCase):
    def setUp(self):
        self.yaml = utils.Yaml()
        self.filename = "test.yaml"
        self.dict = {"foo": "bar"}

    def tearDown(self):
        if os.path.exists(self.filename):
            os.remove(self.filename)

    def test_instantiate_class(self):
        pass

    def test_has_dict_property(self):
        self.assertTrue(hasattr(self.yaml, "dict"))

    def test_dict_property_is_ordered_dict(self):
        self.assertTrue(isinstance(self.yaml.dict, collections.OrderedDict))

    def test_has_read_from_method(self):
        self.assertTrue(hasattr(self.yaml, "read_from"))
        self.assertTrue(callable(self.yaml.read_from))

    def test_read_from_without_filename_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingFilenameError):
            self.yaml.read_from()

    def test_has_write_to_method(self):
        self.assertTrue(hasattr(self.yaml, "write_to"))
        self.assertTrue(callable(self.yaml.write_to))

    def test_write_to_without_filename_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingFilenameError):
            self.yaml.write_to()

    def test_read_from_reads_file(self):
        with open(self.filename, "w") as file:
            yaml.dump(self.dict, file)
        self.yaml.read_from(self.filename)
        self.assertEqual(self.dict, self.yaml.dict)

    def test_writes_to_writes_file(self):
        self.yaml.dict = self.dict
        self.yaml.write_to(self.filename)
        with open(self.filename, "r") as file:
            contents = yaml.load(file, Loader=yaml.SafeLoader)
        self.assertEqual(contents, self.yaml.dict)

    def test_has_serialise_numpy_array_method(self):
        self.assertTrue(hasattr(self.yaml, "serialise_numpy_arrays"))
        self.assertTrue(callable(self.yaml.serialise_numpy_arrays))

    def test_has_serialize_numpy_array_method(self):
        self.assertTrue(hasattr(self.yaml, "serialize_numpy_arrays"))
        self.assertTrue(callable(self.yaml.serialize_numpy_arrays))

    def test_serialise_numpy_array_creates_dict(self):
        array = np.asarray([[0.0, 1.0, 2.0], [1.0, 2.0, 3.0]])
        resulting_dict = {
            "foo": {
                "type": "numpy.ndarray",
                "dtype": str(array.dtype),
                "array": array.tolist(),
            }
        }
        self.yaml.dict = {"foo": array}
        self.yaml.serialise_numpy_arrays()
        self.assertDictEqual(resulting_dict, self.yaml.dict)

    def test_serialise_numpy_array_creates_list(self):
        array = np.asarray([[0.0, 1.0, 2.0], [1.0, 2.0, 3.0]])
        resulting_dict = {"foo": array.tolist()}
        self.yaml.dict = {"foo": array}
        self.yaml.numpy_array_to_list = True
        self.yaml.serialise_numpy_arrays()
        self.assertDictEqual(resulting_dict, self.yaml.dict)

    def test_serialise_large_numpy_array_creates_dict_and_file(self):
        array = np.random.rand(self.yaml.numpy_array_size_threshold + 1)
        resulting_dict = {
            "foo": {
                "type": "numpy.ndarray",
                "dtype": str(array.dtype),
                "file": "foo",
            }
        }
        self.yaml.dict = {"foo": array}
        self.yaml.serialise_numpy_arrays()
        # Necessary as the true filename will be generated
        filename = self.yaml.dict["foo"]["file"]
        resulting_dict["foo"]["file"] = filename
        self.assertDictEqual(resulting_dict, self.yaml.dict)
        self.assertTrue(os.path.exists(filename))
        os.remove(filename)

    def test_serialise_large_numpy_array_sets_filenames(self):
        array = np.random.rand(self.yaml.numpy_array_size_threshold + 1)
        self.yaml.dict = {"foo": array}
        self.yaml.serialise_numpy_arrays()
        # Necessary as the true filename will be generated
        filename = self.yaml.dict["foo"]["file"]
        self.assertTrue(self.yaml.binary_files)
        os.remove(filename)

    def test_serialise_large_non_c_contiguous_numpy_array(self):
        array = np.random.random([6, 8])
        self.yaml.numpy_array_size_threshold = 5
        self.yaml.dict = {"foo": array.T}
        self.yaml.serialise_numpy_arrays()
        # Necessary as the true filename will be generated
        filename = self.yaml.dict["foo"]["file"]
        self.assertTrue(self.yaml.binary_files)
        os.remove(filename)

    def test_serialise_large_numpy_array_with_binary_directory(self):
        array = np.random.rand(self.yaml.numpy_array_size_threshold + 1)
        resulting_dict = {
            "foo": {
                "type": "numpy.ndarray",
                "dtype": str(array.dtype),
                "file": "foo",
            }
        }
        self.yaml.dict = {"foo": array}
        self.yaml.binary_directory = "bar"
        self.yaml.serialise_numpy_arrays()
        # Necessary as the true filename will be generated
        filename = self.yaml.dict["foo"]["file"]
        resulting_dict["foo"]["file"] = filename
        self.assertDictEqual(resulting_dict, self.yaml.dict)
        self.assertTrue(os.path.exists(os.path.join("bar", filename)))
        os.remove(os.path.join("bar", filename))
        os.rmdir("bar")

    def test_serialise_numpy_array_in_hierarchical_dict_creates_dict(self):
        array = np.asarray([[0.0, 1.0, 2.0], [1.0, 2.0, 3.0]])
        array_dict = {
            "foo": {
                "type": "numpy.ndarray",
                "dtype": str(array.dtype),
                "array": array.tolist(),
            }
        }
        resulting_dict = {"bar": array_dict}
        self.yaml.dict = {"bar": {"foo": array}}
        self.yaml.serialise_numpy_arrays()
        self.assertDictEqual(resulting_dict, self.yaml.dict)

    def test_serialise_numpy_array_in_list_in_dict_creates_dict(self):
        array = np.asarray([[0.0, 1.0, 2.0], [1.0, 2.0, 3.0]])
        array_dict = {
            "foo": {
                "type": "numpy.ndarray",
                "dtype": str(array.dtype),
                "array": array.tolist(),
            }
        }
        resulting_dict = {"bar": [array_dict]}
        self.yaml.dict = {"bar": [{"foo": array}]}
        self.yaml.serialise_numpy_arrays()
        self.assertDictEqual(resulting_dict, self.yaml.dict)

    def test_serialise_numpy_array_in_ordered_dict_creates_dict(self):
        array = np.asarray([[0.0, 1.0, 2.0], [1.0, 2.0, 3.0]])
        array_dict = {
            "foo": {
                "type": "numpy.ndarray",
                "dtype": str(array.dtype),
                "array": array.tolist(),
            }
        }
        resulting_dict = {
            "foobar": collections.OrderedDict({"bar": array_dict})
        }
        self.yaml.dict = {
            "foobar": collections.OrderedDict({"bar": {"foo": array}})
        }
        self.yaml.serialise_numpy_arrays()
        self.assertDictEqual(resulting_dict, self.yaml.dict)

    def test_has_deserialise_numpy_array_method(self):
        self.assertTrue(hasattr(self.yaml, "deserialise_numpy_arrays"))
        self.assertTrue(callable(self.yaml.deserialise_numpy_arrays))

    def test_has_deserialize_numpy_array_method(self):
        self.assertTrue(hasattr(self.yaml, "deserialize_numpy_arrays"))
        self.assertTrue(callable(self.yaml.deserialize_numpy_arrays))

    def test_deserialise_numpy_array_creates_numpy_array(self):
        array = np.asarray([[0.0, 1.0, 2.0], [1.0, 2.0, 3.0]])
        resulting_dict = {"foo": array}
        self.yaml.dict = {
            "foo": {
                "type": "numpy.ndarray",
                "dtype": str(array.dtype),
                "array": array.tolist(),
            }
        }
        self.yaml.deserialise_numpy_arrays()
        np.testing.assert_allclose(
            resulting_dict["foo"], self.yaml.dict["foo"]
        )

    def test_deserialise_large_numpy_array_creates_numpy_array(self):
        array = np.random.rand(self.yaml.numpy_array_size_threshold + 1)
        resulting_dict = {"foo": array}
        self.yaml.dict = {
            "foo": {
                "type": "numpy.ndarray",
                "dtype": str(array.dtype),
                "file": "foo.npy",
            }
        }
        np.save("foo", array, allow_pickle=False)
        self.yaml.deserialise_numpy_arrays()
        np.testing.assert_allclose(
            resulting_dict["foo"], self.yaml.dict["foo"]
        )
        os.remove("foo.npy")

    def test_deserialise_large_numpy_array_with_binary_directory(self):
        array = np.random.rand(self.yaml.numpy_array_size_threshold + 1)
        resulting_dict = {"foo": array}
        self.yaml.dict = {
            "foo": {
                "type": "numpy.ndarray",
                "dtype": str(array.dtype),
                "file": "foo.npy",
            }
        }
        self.yaml.binary_directory = "bar"
        os.mkdir("bar")
        np.save("bar/foo.npy", array, allow_pickle=False)
        self.yaml.deserialise_numpy_arrays()
        np.testing.assert_allclose(
            resulting_dict["foo"], self.yaml.dict["foo"]
        )
        os.remove("bar/foo.npy")
        os.rmdir("bar")

    def test_deserialise_numpy_array_in_hierarchical_dict_creates_numpy_array(
        self,
    ):
        array = np.asarray([[0.0, 1.0, 2.0], [1.0, 2.0, 3.0]])
        resulting_dict = {"bar": {"foo": array}}
        self.yaml.dict = {
            "bar": {
                "foo": {
                    "type": "numpy.ndarray",
                    "dtype": str(array.dtype),
                    "array": array.tolist(),
                }
            }
        }
        self.yaml.deserialise_numpy_arrays()
        np.testing.assert_allclose(
            resulting_dict["bar"]["foo"], self.yaml.dict["bar"]["foo"]
        )

    def test_deserialise_numpy_array_in_list_in_dict_creates_numpy_array(
        self,
    ):
        array = np.asarray([[0.0, 1.0, 2.0], [1.0, 2.0, 3.0]])
        resulting_dict = {"bar": [{"foo": array}]}
        self.yaml.dict = {
            "bar": [
                {
                    "foo": {
                        "type": "numpy.ndarray",
                        "dtype": str(array.dtype),
                        "array": array.tolist(),
                    }
                }
            ]
        }
        self.yaml.deserialise_numpy_arrays()
        np.testing.assert_allclose(
            resulting_dict["bar"][0]["foo"], self.yaml.dict["bar"][0]["foo"]
        )

    def test_deserialise_numpy_array_in_ordered_dict_creates_numpy_array(
        self,
    ):
        array = np.asarray([[0.0, 1.0, 2.0], [1.0, 2.0, 3.0]])
        resulting_dict = {
            "foobar": collections.OrderedDict({"bar": {"foo": array}})
        }
        self.yaml.dict = {
            "foobar": collections.OrderedDict(
                {
                    "bar": {
                        "foo": {
                            "type": "numpy.ndarray",
                            "dtype": str(array.dtype),
                            "array": array.tolist(),
                        }
                    }
                }
            )
        }
        self.yaml.deserialise_numpy_arrays()
        np.testing.assert_allclose(
            resulting_dict["foobar"]["bar"]["foo"],
            self.yaml.dict["foobar"]["bar"]["foo"],
        )

    def test_write_stream(self):
        self.yaml.dict = self.dict
        dump = yaml.dump(self.dict)
        self.assertEqual(dump, self.yaml.write_stream())

    def test_write_stream_with_numpy_float(self):
        self.yaml.dict = {"foo": np.float64(5)}
        self.assertEqual("foo: 5.0", self.yaml.write_stream().strip())

    def test_write_stream_with_numpy_int(self):
        self.yaml.dict = {"foo": np.int64(5)}
        self.assertEqual("foo: 5", self.yaml.write_stream().strip())

    def test_read_stream(self):
        self.yaml.dict = self.dict
        dump = yaml.dump(self.dict)
        yaml_object = aspecd.utils.Yaml()
        yaml_object.read_stream(dump)
        self.assertEqual(self.dict, yaml_object.dict)

    def test_read_from_converts_floats_with_exponent_to_float(self):
        yaml_dict = {"foo": "1e3"}
        with open(self.filename, "w") as file:
            yaml.dump(yaml_dict, file)
        self.yaml.read_from(self.filename)
        self.assertEqual(1e3, self.yaml.dict["foo"])

    def test_read_stream_converts_floats_with_exponent_to_float(self):
        yaml_dict = {"foo": "1e3"}
        dump = yaml.dump(yaml_dict)
        yaml_object = aspecd.utils.Yaml()
        yaml_object.read_stream(dump)
        self.assertEqual(1e3, yaml_object.dict["foo"])

    def test_write_stream_converts_tuple_to_list(self):
        yaml_dict = {"foo": (1, 2)}
        self.yaml.dict = yaml_dict
        dump = self.yaml.write_stream()
        self.assertNotIn("!!python/tuple", dump)


class TestReplaceValueInDict(unittest.TestCase):
    def test_replace_value_returns_dict(self):
        target_dict = {"kfoo": "vfoo", "kbar": "vbar"}
        replacement = {"kfoo": "vbar"}
        target_dict = utils.replace_value_in_dict(replacement, target_dict)
        self.assertTrue(isinstance(target_dict, dict))

    def test_replace_value_replaces_existing_value(self):
        target_dict = {"kfoo": "vfoo", "kbar": "vbar"}
        replacement = {"vfoo": "vbar"}
        target_dict = utils.replace_value_in_dict(replacement, target_dict)
        self.assertEqual(replacement["vfoo"], target_dict["kfoo"])

    def test_replace_value_replaces_existing_value_in_cascaded_dict(self):
        target_dict = {"kfoo": {"kbar": "vfoo"}}
        replacement = {"vfoo": "vbar"}
        target_dict = utils.replace_value_in_dict(replacement, target_dict)
        self.assertEqual(replacement["vfoo"], target_dict["kfoo"]["kbar"])

    def test_replace_value_replaces_existing_value_in_doubly_cascaded_dict(
        self,
    ):
        target_dict = {"kfoo": {"kfoobar": {"kbar": "vfoo"}}}
        replacement = {"vfoo": "vbar"}
        target_dict = utils.replace_value_in_dict(replacement, target_dict)
        self.assertEqual(
            replacement["vfoo"], target_dict["kfoo"]["kfoobar"]["kbar"]
        )

    def test_replace_value_ignores_nonexisting_value(self):
        target_dict = {"kfoo": "vfoo", "kbar": "vbar"}
        replacement = {"kfoo": "vbar"}
        modified_dict = utils.replace_value_in_dict(replacement, target_dict)
        self.assertEqual(modified_dict, target_dict)

    def test_replace_value_ignores_nonexisting_value_in_cascaded_dict(self):
        target_dict = {"kfoo": {"kbar": "vfoo"}}
        replacement = {"kfoo": "vbar"}
        modified_dict = utils.replace_value_in_dict(replacement, target_dict)
        self.assertEqual(modified_dict, target_dict)

    def test_replace_value_replaces_existing_value_in_list(self):
        target_dict = {"kfoo": ["vfoo", "vbar"]}
        replacement = {"vfoo": "vbar"}
        target_dict = utils.replace_value_in_dict(replacement, target_dict)
        self.assertEqual(replacement["vfoo"], target_dict["kfoo"][0])

    def test_replace_value_handles_lists_of_dicts(self):
        target_dict = {"kfoo": [{"kfoo": "vfoo"}, {"kbar": "vbar"}]}
        replacement = {"vfoo": "vbar"}
        target_dict = utils.replace_value_in_dict(replacement, target_dict)
        self.assertEqual(replacement["vfoo"], target_dict["kfoo"][0]["kfoo"])

    def test_replace_value_replaces_empty_list_in_target(self):
        target_dict = {"kfoo": []}
        replacement = {"kfoo": ["foo", "bar"]}
        target_dict = utils.replace_value_in_dict(replacement, target_dict)
        self.assertEqual(replacement["kfoo"], target_dict["kfoo"])


class TestCopyValuesBetweenDicts(unittest.TestCase):
    def test_copy_values_returns_dict(self):
        target_dict = utils.copy_values_between_dicts(dict(), dict())
        self.assertTrue(isinstance(target_dict, dict))

    def test_copy_values_copies_existing_value(self):
        source = {"kfoo": "vbar"}
        target = {"kfoo": "vfoo", "kbar": "vbar"}
        target = utils.copy_values_between_dicts(source, target)
        self.assertEqual(source["kfoo"], target["kfoo"])

    def test_copy_values_ignores_nonexisting_value(self):
        source = {"kfoobar": "vbar"}
        target = {"kfoo": "vfoo", "kbar": "vbar"}
        modified_target = utils.copy_values_between_dicts(source, target)
        self.assertEqual(modified_target, target)

    def test_copy_values_copies_existing_value_in_cascaded_target_dict(self):
        source = {"kbar": "vfoo"}
        target = {"kfoo": {"kbar": "vbar"}}
        target = utils.copy_values_between_dicts(source, target)
        self.assertEqual(source["kbar"], target["kfoo"]["kbar"])

    def test_copy_values_copies_existing_value_in_cascaded_dicts(self):
        source = {"kfoo": {"kbar": "vfoo"}}
        target = {"kfoo": {"kbar": "vbar"}}
        target = utils.copy_values_between_dicts(source, target)
        self.assertEqual(source["kfoo"]["kbar"], target["kfoo"]["kbar"])


class TestCopyKeysBetweenDicts(unittest.TestCase):
    def test_copy_keys_returns_dict(self):
        target_dict = utils.copy_keys_between_dicts(dict(), dict())
        self.assertTrue(isinstance(target_dict, dict))

    def test_copy_keys_copies_existing_key(self):
        source = {"kfoo": "vbar"}
        target = {"kfoo": "vfoo", "kbar": "vbar"}
        target = utils.copy_keys_between_dicts(source, target)
        self.assertEqual(source["kfoo"], target["kfoo"])

    def test_copy_keys_copies_existing_key_in_cascaded_dicts(self):
        source = {"kfoo": {"kbar": "vfoo"}}
        target = {"kfoo": {"kbar": "vbar"}}
        target = utils.copy_keys_between_dicts(source, target)
        self.assertEqual(source["kfoo"]["kbar"], target["kfoo"]["kbar"])

    def test_copy_keys_preserves_existing_keys_in_target_dicts(self):
        source = {"kfoo": {"kbar": "vfoo"}}
        target = {"kfoo": {"kbar": "vbar", "kfoobar": "vfoobar"}}
        target = utils.copy_keys_between_dicts(source, target)
        self.assertIn("kfoobar", target["kfoo"])

    def test_copy_keys_of_cascaded_dict_with_nonexisting_target_key(self):
        source = {"kfoo": {"kbar": "vfoo"}}
        target = {"kbar": {"kbar": "vbar", "kfoobar": "vfoobar"}}
        target = utils.copy_keys_between_dicts(source, target)
        self.assertIn("kfoo", target)

    def test_copy_keys_if_dict_in_list(self):
        source = {"kfoo": [{"kbar": "vfoo"}]}
        target = {"kfoo": [{"kbar": "vbar", "kfoobar": "vfoobar"}]}
        target = utils.copy_keys_between_dicts(source, target)
        self.assertEqual("vfoo", target["kfoo"][0]["kbar"])

    def test_copy_key_if_target_key_is_dict_but_source_key_is_not(self):
        source = {"kfoo": "kbar"}
        target = {"kfoo": {"kbar": "vfoo"}}
        target = utils.copy_keys_between_dicts(source, target)
        self.assertEqual(source["kfoo"], target["kfoo"])


class TestRemoveEmptyValuesFromDicts(unittest.TestCase):
    def test_remove_empty_values_returns_dict(self):
        target_dict = utils.remove_empty_values_from_dict(dict())
        self.assertTrue(isinstance(target_dict, dict))

    def test_remove_empty_values_removes_keys_with_empty_values(self):
        dict_ = {"foo": ""}
        target_dict = utils.remove_empty_values_from_dict(dict_)
        self.assertEqual(target_dict, dict())

    def test_remove_empty_values_only_removes_keys_with_empty_values(self):
        dict_ = {"foo": "", "bar": "bla"}
        target_dict = utils.remove_empty_values_from_dict(dict_)
        self.assertEqual(target_dict, {"bar": "bla"})

    def test_remove_empty_values_recursively_removes_keys(self):
        dict_ = {"foo": {"bar": ""}}
        target_dict = utils.remove_empty_values_from_dict(dict_)
        self.assertEqual(target_dict, dict())

    def test_remove_empty_values_removes_deep_keys(self):
        dict_ = {"foo": {"bar": "", "bla": "blub"}}
        target_dict = utils.remove_empty_values_from_dict(dict_)
        self.assertEqual(target_dict, {"foo": {"bla": "blub"}})


class TestConvertKeysToVariableNames(unittest.TestCase):
    def test_convert_keys_returns_dict(self):
        target_dict = utils.convert_keys_to_variable_names(dict())
        self.assertTrue(isinstance(target_dict, dict))

    def test_convert_keys_converts_to_lower_case(self):
        dict_ = {"Foo": "bar"}
        target_dict = utils.convert_keys_to_variable_names(dict_)
        self.assertDictEqual({"foo": "bar"}, target_dict)

    def test_convert_keys_converts_space_to_underscore(self):
        dict_ = {"foo bar": "bar"}
        target_dict = utils.convert_keys_to_variable_names(dict_)
        self.assertDictEqual({"foo_bar": "bar"}, target_dict)

    def test_convert_keys_recursively_converts_to_lower_case(self):
        dict_ = {"Foo": {"Bar": "bla"}}
        target_dict = utils.convert_keys_to_variable_names(dict_)
        self.assertDictEqual({"foo": {"bar": "bla"}}, target_dict)

    def test_convert_keys_recursively_converts_space_to_underscore(self):
        dict_ = {"foo bar": {"bar bla": "blub"}}
        target_dict = utils.convert_keys_to_variable_names(dict_)
        self.assertDictEqual({"foo_bar": {"bar_bla": "blub"}}, target_dict)


class TestAllEqual(unittest.TestCase):
    def test_equal_elements_of_list_return_true(self):
        self.assertTrue(utils.all_equal(["foo", "foo"]))

    def test_different_elements_of_list_return_false(self):
        self.assertFalse(utils.all_equal(["foo", "bar"]))


class TestProperties(unittest.TestCase):
    def setUp(self):
        self.properties = aspecd.utils.Properties()

    def test_instantiate_class(self):
        pass

    def test_has_to_dict_method(self):
        self.assertTrue(hasattr(self.properties, "to_dict"))
        self.assertTrue(callable(self.properties.to_dict))

    def test_has_from_dict_method(self):
        self.assertTrue(hasattr(self.properties, "from_dict"))
        self.assertTrue(callable(self.properties.from_dict))

    def test_from_dict_without_dict_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingDictError):
            self.properties.from_dict()

    def test_from_dict_sets_attribute(self):
        attribute = "foo"
        dict_ = dict()
        dict_[attribute] = "foo"
        setattr(self.properties, attribute, None)
        self.properties.from_dict(dict_)
        self.assertEqual(
            dict_[attribute], getattr(self.properties, attribute)
        )

    def test_from_dict_with_list_attribute_appends_to_list(self):
        attribute = "foo"
        dict_ = dict()
        dict_[attribute] = "foo"
        setattr(self.properties, attribute, ["bar"])
        self.properties.from_dict(dict_)
        self.assertEqual(
            ["bar", dict_[attribute]], getattr(self.properties, attribute)
        )

    def test_from_dict_with_list_attribute_appends_each_element(self):
        attribute = "foo"
        dict_ = dict()
        dict_[attribute] = ["foo", "bar"]
        setattr(self.properties, attribute, ["bar"])
        self.properties.from_dict(dict_)
        self.assertEqual(
            ["bar", "foo", "bar"], getattr(self.properties, attribute)
        )

    def test_from_dict_does_not_set_unknown_attribute(self):
        attribute = "foo"
        dict_ = dict()
        dict_[attribute] = "foo"
        self.properties.from_dict(dict_)
        self.assertFalse(hasattr(self.properties, attribute))

    def test_from_dict_with_property_object_as_property(self):
        dict_ = {"prop1": {"prop2": "foo"}}
        prop1 = aspecd.utils.Properties()
        setattr(prop1, "prop2", "")
        setattr(self.properties, "prop1", prop1)
        self.properties.from_dict(dict_)
        # noinspection PyUnresolvedReferences
        self.assertEqual("foo", self.properties.prop1.prop2)

    def test_has_get_properties_method(self):
        self.assertTrue(hasattr(self.properties, "get_properties"))
        self.assertTrue(callable(self.properties.get_properties))

    def test_get_properties_without_properties_returns_empty_list(self):
        self.assertEqual([], self.properties.get_properties())

    def test_get_properties_returns_list_of_properties(self):
        props = ["foo", "bar"]
        for prop in props:
            setattr(self.properties, prop, None)
        self.assertEqual(props, self.properties.get_properties())

    def test_get_properties_does_not_return_nonpublic_properties(self):
        props = ["_foo", "bar"]
        for prop in props:
            setattr(self.properties, prop, None)
        self.assertEqual(["bar"], self.properties.get_properties())

    def test_get_properties_does_not_return_excluded_properties(self):
        props = ["foo", "bar"]
        for prop in props:
            setattr(self.properties, prop, None)
        self.properties._exclude = ["foo"]
        self.assertEqual(["bar"], self.properties.get_properties())


class TestBasename(unittest.TestCase):
    def test_basename_returns_last_part_of_path_without_extension(self):
        self.assertEqual("bar", utils.basename("/foo/bar.bla"))

    def test_basename_returns_last_part_if_starting_with_dot(self):
        self.assertEqual(".bashrc", utils.basename("/home/doe/.bashrc"))

    def test_basename_removes_only_last_extension(self):
        self.assertEqual("foo.tar", utils.basename("/home/doe/foo.tar.gz"))

    def test_basename_works_without_path(self):
        self.assertEqual("foo", utils.basename("foo.bar"))

    def test_basename_works_without_extension(self):
        self.assertEqual("foo", utils.basename("/path/to/foo"))

    def test_basename_works_with_relative_path(self):
        self.assertEqual("foo", utils.basename("./path/to/foo"))


class TestPath(unittest.TestCase):
    def test_path_returns_path_without_filename(self):
        self.assertEqual("/foo/", utils.path("/foo/bar.bla"))

    def test_path_works_with_relative_path(self):
        self.assertEqual("./foo/", utils.path("./foo/bar.bla"))


class TestNotZero(unittest.TestCase):
    def test_not_zero_of_zero_returns_nonzero_value(self):
        self.assertGreater(utils.not_zero(0), 0)

    def test_not_zero_of_zero_returns_np_float_resolution(self):
        self.assertEqual(np.finfo(np.float64).resolution, utils.not_zero(0))

    def test_not_zero_of_positive_value_preserves_sign(self):
        self.assertGreater(utils.not_zero(1e-20), 0)

    def test_not_zero_of_negative_value_preserves_sign(self):
        self.assertLess(utils.not_zero(-1e-20), 0)

    def test_not_zero_of_negative_value_closer_than_limit_returns_limit(self):
        self.assertEqual(
            -np.finfo(np.float64).resolution, utils.not_zero(-1e-20)
        )


class TestIterable(unittest.TestCase):
    def test_iterable_returns_true_for_list(self):
        self.assertTrue(utils.isiterable([]))

    def test_iterable_returns_true_for_tuple(self):
        self.assertTrue(utils.isiterable(tuple()))

    def test_iterable_returns_true_for_nd_array(self):
        self.assertTrue(utils.isiterable(np.asarray([])))

    def test_iterable_returns_true_for_string(self):
        self.assertTrue(utils.isiterable("foo"))

    def test_iterable_returns_false_for_scalar(self):
        self.assertFalse(utils.isiterable(1))


class TestGetPackageData(unittest.TestCase):
    def setUp(self):
        self.filename = "bar"
        self.data_dir = "foo"

    def tearDown(self):
        if os.path.exists(self.data_dir):
            shutil.rmtree(self.data_dir)

    def create_data_dir_and_contents(self):
        data_dir = os.path.join(self.data_dir, "aspecd")
        os.makedirs(data_dir)
        with open(
            os.path.join(data_dir, self.filename), "w+", encoding="utf8"
        ) as file:
            file.write("foo")

    def test_get_package_data_without_name_raises(self):
        with self.assertRaises(ValueError):
            utils.get_package_data()

    def test_get_package_data_returns_content(self):
        with patch("pkgutil.get_data", return_value="foo".encode()):
            content = utils.get_package_data(self.filename)
        self.assertTrue(content)

    def test_get_package_data_with_prefixed_package_returns_content(self):
        content = utils.get_package_data("aspecd@utils.py")
        self.assertTrue(content)

    def test_get_package_data_with_foreign_package_returns_content(self):
        content = utils.get_package_data("pip@__main__.py", directory="")
        self.assertTrue(content)


class TestChangeWorkingDir(unittest.TestCase):
    def test_change_working_dir_changes_working_dir(self):
        with utils.change_working_dir(".."):
            working_dir = os.path.abspath(os.getcwd())
        self.assertEqual(os.path.split(os.getcwd())[0], working_dir)

    def test_change_working_dir_returns_to_original_dir(self):
        oldpwd = os.getcwd()
        with utils.change_working_dir(".."):
            pass
        self.assertEqual(oldpwd, os.getcwd())


class TestGetLogger(unittest.TestCase):
    def test_get_logger_returns_logger(self):
        logger = utils.get_logger()
        self.assertIsInstance(logger, logging.Logger)

    def test_get_logger_without_argument_returns_package_logger(self):
        logger = utils.get_logger()
        self.assertEqual(utils.package_name(), logger.name)

    def test_get_logger_with_argument_returns_child_logger(self):
        logger = utils.get_logger(__name__)
        name = ".".join([utils.package_name(), __name__])
        self.assertEqual(name, logger.name)

    def test_get_logger_returns_logger_with_null_handler(self):
        logger = utils.get_logger(__name__)
        self.assertIsInstance(logger.handlers[0], logging.NullHandler)



"""Obtaining information on the system used for data processing and analysis.

One key aspect of reproducibility is to record sufficient details of the
system used to perform processing and analysis. Therefore, each
:class:`aspecd.dataset.HistoryRecord` contains a field with system
information that is an :obj:`aspecd.system.SystemInfo` object.

General information stored within the :class:`aspecd.system.SystemInfo`
class are the Python version, the platform, as well as the login name of the
user currently logged in. Therefore, this is a relevant aspect for personal
data protection, and each and every user of the system should be made
available of this fact.

"""

import getpass
import platform
import sys

import pkg_resources

import aspecd.utils


class SystemInfo(aspecd.utils.ToDictMixin):
    """
    General information on the system used.

    Attributes
    ----------
    python : :class:`dict`
        Version of Python (and potentially further information)
    packages : :class:`dict`
        Relevant modules and their version numbers
    platform : :class:`string`
        Identifier of the platform
    user : :class:`dict`
        Currently only the login name of the currently logged-in user

    Parameters
    ----------
    package : :class:`str`
        Name of package whose version shall be added to the :attr:`modules`
        dictionary

        Useful (and necessary) for packages derived from the ASpecD
        framework to store their version number in the SystemInfo class and
        hence in the history records. Prerequisite for reproducibility.

    """

    def __init__(self, package=""):
        super().__init__()
        self.python = {}
        self.packages = {}
        self.platform = platform.platform()
        self.user = {}
        # Set some properties of dicts
        self._set_values(package)

    def _set_values(self, package):
        self.python["version"] = sys.version
        self.packages["aspecd"] = aspecd.utils.get_aspecd_version()
        self._add_requirements_to_packages(package="aspecd")
        if package and package != "aspecd":
            self.packages[package] = aspecd.utils.package_version(package)
            self._add_requirements_to_packages(package=package)
        self.user["login"] = getpass.getuser()

    def _add_requirements_to_packages(self, package="aspecd"):
        requirements = [
            requirement.name
            for requirement in pkg_resources.get_distribution(
                package
            ).requires()
        ]
        for requirement in requirements:
            self.packages[requirement] = pkg_resources.get_distribution(
                requirement
            ).version

    def from_dict(self, dict_=None):
        """
        Set properties from dictionary.

        Only parameters in the dictionary that are valid properties of the
        class are set accordingly.

        Parameters
        ----------
        dict_ : :class:`dict`
            Dictionary containing properties to set

        """
        for key, value in dict_.items():
            if hasattr(self, key):
                if isinstance(value, dict):
                    for sub_key, sub_value in value.items():
                        getattr(self, key)[sub_key] = sub_value
                else:
                    setattr(self, key, value)



"""Datasets: units containing data and metadata.

The dataset is one :doc:`key concept <../concepts>` of the ASpecD framework,
consisting of the data as well as the corresponding metadata. Storing
metadata in a structured way is a prerequisite for a semantic
understanding within the routines. Furthermore, a history of every
processing, analysis and annotation step is recorded as well, aiming at a
maximum of reproducibility. This is part of how the ASpecD framework tries
to support good scientific practice.

Therefore, each processing and analysis step of data should always be
performed using the respective methods of a dataset, at least as long as it
can be performed on a single dataset.


Types of datasets
=================

Generally, there are two types of datasets: Those containing experimental
data and those containing calculated data. Therefore, two corresponding
subclasses exist, and packages building upon the ASpecD framework should
inherit from either of them:

  * :class:`aspecd.dataset.ExperimentalDataset`
  * :class:`aspecd.dataset.CalculatedDataset`

Calculated datasets can either be the result of actual simulations or
dummy datasets used for testing purposes.


Classes used by the dataset
===========================

Additional classes used within the dataset that are normally not necessary
to implement directly on your own in packages building upon the ASpecD
framework, are:

  * :class:`aspecd.dataset.Data`

    Unit containing both, numeric data and corresponding axes.

    The data class ensures consistency in terms of dimensions between
    numerical data and axes.

  * :class:`aspecd.dataset.Axis`

    Axis for data in a dataset.

    An axis contains always both, numerical values as well as the metadata
    necessary to create axis labels and to make sense of the numerical
    information.

  * :class:`aspecd.dataset.DeviceData`

    Additional data from devices recorded parallel to the actual data.

    The dataset concept (see :class:`aspecd.dataset.Dataset`) rests on the
    assumption that there is one particular set of data that can be
    regarded as the actual or primary data of the dataset. However,
    in many cases, parallel to these actual data, other data are recorded
    as well, be it readouts from monitors or alike.

  * :class:`aspecd.dataset.DatasetReference`

    Reference to a dataset.

    Often, one dataset needs to reference other datasets. A typical example
    would be a simulation stored in a dataset of class
    :class:`aspecd.dataset.CalculatedDataset` that needs to reference the
    corresponding experimental data, stored in a dataset of class
    :class:`aspecd.dataset.ExperimentalDataset`. Vice versa,
    the experimental dataset might want to store a reference to one (or
    more) simulations.


.. _sec:dataset:device_data:

Device data
===========

The dataset concept (see :class:`aspecd.dataset.Dataset`) rests on the
assumption that there is one particular set of data that can be
regarded as the actual or primary data of the dataset. However,
in many cases, parallel to these actual data, other data are recorded
as well, be it readouts from monitors or alike.

Usually, these additional data will share one axis with the
primary data of the dataset. However, this is not necessarily the
case. Furthermore, one dataset may contain an arbitrary number of
additional device data entries.

Technically speaking, :class:`aspecd.dataset.DeviceData` are a special
or extended type of :class:`aspecd.dataset.Data`, *i.e.* a unit
containing both numerical data and corresponding axes. However,
this class extends that with metadata specific for the device the
additional data have been recorded with. Why storing metadata here and
not in the :attr:`aspecd.dataset.Dataset.metadata` property? The
latter is more concerned with an overall description of the
experimental setup in sufficient detail, while the metadata contained
in this class are more device-specific. Potential contents of the
metadata here are internal device IDs, addresses for communication,
and alike. Eventually, the metadata contained herein are those that
can be relevant mainly for debugging purposes or sanity checks of
experiments.

.. admonition:: Example

    A real example for additional data recorded in spectroscopy comes
    from time-resolved EPR (tr-EPR) spectroscopy: Here, you usually
    record 2D data as function of magnetic field and time, *i.e.* a
    full time profile per magnetic field point. As this is a
    non-standard method, often setups are controlled by lab-written
    software and allow for monitoring parameters not usually recorded
    with commercial setups. In this particular case, this can be the
    time stamp and microwave frequency for each individual recorded
    time trace, and the `Python trEPR package
    <https://docs.trepr.de/>`_ not only handles tr-EPR data, but is
    capable of dealing with both additional types of data for analysis
    purposes.


Module documentation
====================

"""

import copy

import numpy as np

import aspecd.exceptions
import aspecd.history
import aspecd.io
import aspecd.metadata
import aspecd.system
import aspecd.utils


class Dataset(aspecd.utils.ToDictMixin):
    """Base class for all kinds of datasets.

    The dataset is one of the core elements of the ASpecD framework, basically
    containing both, (numeric) data and corresponding metadata, aka information
    available about the data.

    Generally, there are two types of datasets: Those containing
    experimental data and those containing calculated data. Therefore,
    two corresponding subclasses exist, and packages building upon the
    ASpecD framework should inherit from either of them:

      * :class:`aspecd.dataset.ExperimentalDataset`
      * :class:`aspecd.dataset.CalculatedDataset`

    The public attributes of a dataset can be converted to a dict via
    :meth:`aspecd.utils.ToDictMixin.to_dict()`.

    Attributes
    ----------
    id : :class:`str`
        (unique) identifier of the dataset (i.e., path, LOI, or else)

    label : :class:`str`
        Short description of the dataset

        Can be set by the user, defaults to the value set as
        :attr:`aspecd.dataset.id` by the importer.

    data : :obj:`aspecd.dataset.Data`
        numeric data and axes

    device_data : :class:`dict`
        Additional data from devices recorded parallel to the actual data.

        For details and a bit of background, see
        :class:`aspecd.dataset.DeviceData`.

        In a real dataset with actual device data, the key typically
        identifies the device (type) and the value is of type
        :class:`aspecd.dataset.DeviceData`.

        .. note::

            To further process and analyse these device data, the most
            general way is to extract them as individual dataset each and
            perform all further tasks on it, respectively. See
            :class:`aspecd.analysis.DeviceDataExtraction` for details.

        .. versionadded:: 0.9

    metadata : :obj:`aspecd.metadata.DatasetMetadata`
        hierarchical key-value store of metadata

    history : :class:`list`
        processing steps performed on the numeric data

        For a full list of tasks performed on a dataset in *chronological*
        order see the :attr:`aspecd.dataset.Dataset.tasks` attribute.

    analyses : :class:`list`
        analysis steps performed on the dataset

        For a full list of tasks performed on a dataset in *chronological*
        order see the :attr:`aspecd.dataset.Dataset.tasks` attribute.

    annotations : :class:`list`
        annotations of the dataset

        For a full list of tasks performed on a dataset in *chronological*
        order see the :attr:`aspecd.dataset.Dataset.tasks` attribute.

    representations : :class:`list`
        representations of the dataset, e.g., plots

        For a full list of tasks performed on a dataset in *chronological*
        order see the :attr:`aspecd.dataset.Dataset.tasks` attribute.

    references : :class:`list`
        references to other datasets

        Each reference is an object of type
        :class:`aspecd.dataset.DatasetReference`.

    tasks : :class:`list`
        tasks performed on the dataset in *chronological* order

        Each entry in the list is a dict containing information about the
        type of task (*i.e.*, processing, analysis, annotation,
        representation) and a reference to the object containing more
        information about the respective task.

        Tasks come in quite handy in cases where the exact chronological order
        of steps performed on a dataset are of relevance, regardless of
        their particular type, *e.g.*, in context of reports.

    Raises
    ------
    aspecd.exceptions.UndoWithEmptyHistoryError
        Raised when trying to undo with empty history
    aspecd.exceptions.UndoAtBeginningOfHistoryError
        Raised when trying to undo with history pointer at zero
    aspecd.exceptions.UndoStepUndoableError
        Raised when trying to undo an undoable step of history
    aspecd.exceptions.RedoAlreadyAtLatestChangeError
        Raised  when trying to redo with empty history
    aspecd.exceptions.ProcessingWithLeadingHistoryError
        Raised  when trying to process with leading history

    """

    def __init__(self):
        super().__init__()
        self.data = Data()
        self._origdata = Data()
        self.device_data = {}
        self.metadata = aspecd.metadata.DatasetMetadata()
        self.history = []
        self._history_pointer = -1
        self.analyses = []
        self.annotations = []
        self.representations = []
        self.id = ""  # pylint: disable=invalid-name
        self.label = ""
        self.references = []
        self.tasks = []
        # Package name is used to store the package version in history records
        self._package_name = aspecd.utils.package_name(self)
        self._include_in_to_dict = [
            "_origdata",
            "_package_name",
            "_history_pointer",
        ]

    @property
    def package_name(self):
        """Return package name.

        The name of the package the dataset is implemented in is a crucial
        detail for writing the history. The value is set automatically and
        is read-only.

        """
        return self._package_name

    def process(self, processing_step=None):
        """Apply processing step to dataset.

        Every processing step is an object of type
        :class:`aspecd.processing.SingleProcessingStep` and is passed as
        argument to :meth:`process`.

        Calling this function ensures that the history record is added to the
        dataset as well as a few basic checks are performed such as for leading
        history, meaning that the ``_history_pointer`` is not set to the
        current tip of the history of the dataset. In this case, an error is
        raised.

        .. note::
            If processing_step is undoable, all previous plots stored in
            the list of representations will be removed, as these plots
            cannot be reproduced due to a change in :attr:`_origdata`.

        Parameters
        ----------
        processing_step : :obj:`aspecd.processing.SingleProcessingStep`
            processing step to apply to the dataset

        Returns
        -------
        processing_step : :obj:`aspecd.processing.SingleProcessingStep`
            processing step applied to the dataset

        Raises
        ------
        aspecd.exceptions.ProcessingWithLeadingHistoryError
            Raised when trying to process with leading history

        """
        self._check_processing_prerequisites(processing_step=processing_step)
        # Important: Need a copy, not the reference to the original object
        processing_step = copy.deepcopy(processing_step)
        processing_step.process(self, from_dataset=True)
        history_record = processing_step.create_history_record()
        self.append_history_record(history_record)
        self._append_task(kind="processing", task=history_record)
        self._handle_not_undoable(processing_step=processing_step)
        return processing_step

    def _check_processing_prerequisites(self, processing_step=None):
        if self._has_leading_history():
            raise aspecd.exceptions.ProcessingWithLeadingHistoryError
        if not processing_step:
            raise aspecd.exceptions.MissingProcessingStepError

    def _handle_not_undoable(self, processing_step=None):
        if not processing_step.undoable:
            self._origdata = copy.deepcopy(self.data)
            self.representations = []

    def undo(self):
        """Revert last processing step.

        Actually, the history pointer is decremented and starting from the
        ``_origdata``, all processing steps are reapplied to the data up to
        this point in history.

        Raises
        ------
        aspecd.exceptions.UndoWithEmptyHistoryError
            Raised when trying to undo with empty history
        aspecd.exceptions.UndoAtBeginningOfHistoryError
            Raised when trying to undo with history pointer at zero
        aspecd.exceptions.UndoStepUndoableError
            Raised when trying to undo an undoable step of history

        """
        self._check_undo_prerequisites()
        self._decrement_history_pointer()
        self._replay_history()

    def _check_undo_prerequisites(self):
        if not self.history:
            raise aspecd.exceptions.UndoWithEmptyHistoryError
        if self._history_pointer == -1:
            raise aspecd.exceptions.UndoAtBeginningOfHistoryError
        if self.history[self._history_pointer].undoable:
            raise aspecd.exceptions.UndoStepUndoableError

    def redo(self):
        """Reapply previously undone processing step.

        Raises
        ------
        aspecd.exceptions.RedoAlreadyAtLatestChangeError
            Raised when trying to redo with empty history

        """
        if self._at_tip_of_history():
            raise aspecd.exceptions.RedoAlreadyAtLatestChangeError
        processing_step_record = self.history[
            self._history_pointer + 1
        ].processing
        processing_step = processing_step_record.create_processing_step()
        processing_step.process(self, from_dataset=True)
        self._increment_history_pointer()

    def _at_tip_of_history(self):
        return self._history_pointer == len(self.history) - 1

    def _has_leading_history(self):
        return len(self.history) - 1 > self._history_pointer

    def append_history_record(self, history_record):
        """Append history record to dataset history.

        This method should never be called manually, but only from within
        classes of the ASpecD framework, at least as long as you are not
        interested in Orwellian History.

        Parameters
        ----------
        history_record : :class:`aspecd.history.HistoryRecord`
            History record (of a processing step) to be appended.


        .. versionchanged:: 0.2
            Converted into a public method, due to needs of
            :class:`aspecd.processing.MultiProcessingStep`

        """
        self.history.append(history_record)
        self._increment_history_pointer()

    def _append_task(self, kind="", task=None):
        task = {
            "kind": kind,
            "task": task,
        }
        self.tasks.append(task)

    def _increment_history_pointer(self):
        self._history_pointer += 1

    def _decrement_history_pointer(self):
        self._history_pointer -= 1

    def _replay_history(self):
        self.data = self._origdata
        for history_entry in self.history[: self._history_pointer]:
            history_entry.replay(self)

    def strip_history(self):
        """Remove leading history, if any.

        If a dataset has a leading history, i.e., its history pointer does not
        point to the last entry of the history, and you want to perform a
        processing step on this very dataset, you need first to strip its
        history, as otherwise, a :class:`ProcessingWithLeadingHistoryError`
        will be raised.

        """
        if not self._has_leading_history():
            return
        del self.history[self._history_pointer + 1 :]

    def analyse(self, analysis_step=None):
        """Apply analysis to dataset.

        Every analysis step is an object of type
        :class:`aspecd.analysis.SingleAnalysisStep` and is passed as an argument
        to :meth:`analyse`.

        The information necessary to reproduce an analysis is stored in the
        :attr:`analyses` attribute as object of class
        :class:`aspecd.dataset.AnalysisHistoryRecord`. This record contains as
        well a (deep) copy of the complete history of the dataset stored in
        :attr:`history`.

        Parameters
        ----------
        analysis_step : :obj:`aspecd.analysis.SingleAnalysisStep`
            analysis step to apply to the dataset

        Returns
        -------
        analysis_step : :obj:`aspecd.analysis.SingleAnalysisStep`
            analysis step applied to the dataset

        """
        # Important: Need a copy, not the reference to the original object
        analysis_step = copy.deepcopy(analysis_step)
        analysis_step.analyse(self, from_dataset=True)
        history_record = analysis_step.create_history_record()
        self.analyses.append(history_record)
        self._append_task(kind="analysis", task=history_record)
        return analysis_step

    def analyze(self, analysis_step=None):
        """Apply analysis to dataset.

        Same method as :meth:`analyse`, but for those preferring AE
        over BE.

        """
        analysis_step = self.analyse(analysis_step)
        return analysis_step

    def delete_analysis(self, index=None):
        """Remove analysis step record from dataset.

        Parameters
        ----------
        index : `int`
            Number of analysis in analyses to delete

        """
        del self.analyses[index]

    def annotate(self, annotation_=None):
        """Add annotation to dataset.

        Parameters
        ----------
        annotation_ : :obj:`aspecd.annotation.DatasetAnnotation`
            annotation to add to the dataset

        """
        # Important: Need a copy, not the reference to the original object
        annotation_ = copy.deepcopy(annotation_)
        annotation_.annotate(self, from_dataset=True)
        history_record = annotation_.create_history_record()
        self.annotations.append(history_record)
        self._append_task(kind="annotation", task=history_record)

    def delete_annotation(self, index=None):
        """Remove annotation record from dataset.

        Parameters
        ----------
        index : `int`
            Number of analysis in analyses to delete

        """
        del self.annotations[index]

    def plot(self, plotter=None):
        """Perform plot with data of current dataset.

        Every plotter is an object of type :class:`aspecd.plotting.Plotter`
        and is passed as an argument to :meth:`plot`.

        The information necessary to reproduce a plot is stored in the
        :attr:`representations` attribute as object of class
        :class:`aspecd.dataset.PlotHistoryRecord`. This record contains as
        well a (deep) copy of the complete history of the dataset stored in
        :attr:`history`. Besides being a necessary prerequisite to
        reproduce a plot, this allows to automatically recreate plots
        requiring different incompatible preprocessing steps in arbitrary
        order.

        Parameters
        ----------
        plotter : :obj:`aspecd.plotting.Plotter`
            plot to perform with data of current dataset

        Returns
        -------
        plotter : :obj:`aspecd.plotting.Plotter`
            plot performed on the current dataset

        Raises
        ------
        aspecd.exceptions.MissingPlotterError
            Raised when trying to plot without plotter

        """
        if not plotter:
            raise aspecd.exceptions.MissingPlotterError
        plotter.plot(dataset=self, from_dataset=True)
        plot_record = plotter.create_history_record()
        self.representations.append(plot_record)
        self._append_task(kind="representation", task=plot_record)
        return plotter

    def tabulate(self, table=None):
        """Create table from data of current dataset.

        Every table is an object of type :class:`aspecd.table.Table`
        and is passed as an argument to :meth:`tabulate`.

        The information necessary to reproduce a table is stored in the
        :attr:`representations` attribute as object of class
        :class:`aspecd.dataset.TableHistoryRecord`.

        Parameters
        ----------
        table : :obj:`aspecd.table.Table`
            table created from the data of the current dataset

        Returns
        -------
        table : :obj:`aspecd.table.Table`
            table created from the data of the current dataset

        Raises
        ------
        TypeError
            Raised when trying to tabulate without table

        """
        if not table:
            raise TypeError("tabulate needs a Table object")
        table.tabulate(dataset=self, from_dataset=True)
        table_record = table.create_history_record()
        self.representations.append(table_record)
        self._append_task(kind="representation", task=table_record)
        return table

    def delete_representation(self, index=None):
        """Remove representation record from dataset.

        Parameters
        ----------
        index : `int`
            Number of analysis in analyses to delete

        """
        del self.representations[index]

    def load(self, filename=None):
        """Load dataset object from persistence layer.

        The dataset will be loaded from a file conforming to the ASpecD
        dataset format (adf). For details, see the
        :class:`aspecd.io.AdfExporter` class.

        """
        importer = aspecd.io.AdfImporter()
        importer.source = filename
        importer.import_into(self)

    def save(self, filename=None):
        """Save dataset to persistence layer.

        The dataset will be saved in ASpecD dataset format (adf). For
        details, see the :class:`aspecd.io.AdfExporter` class.

        """
        exporter = aspecd.io.AdfExporter()
        exporter.target = filename
        exporter.export_from(self)

    def import_from(self, importer=None):
        """Import data and metadata contained in importer object.

        This requires initialising an :obj:`aspecd.io.Importer` object
        first that is provided as an argument for this method.

        .. note::
            The same operation can be performed by calling the
            :meth:`import_into` method of an :obj:`aspecd.io.Importer`
            object taking an :obj:`aspecd.dataset.Dataset` object as argument.

            However, as usually one wants to continue working with a dataset,
            first creating an instance of a dataset and a respective importer
            and then calling :meth:`import_from` of the dataset is the
            preferred way.

        Parameters
        ----------
        importer : :class:`aspecd.io.DatasetImporter`
            Importer containing data and metadata read from some source

        """
        if not importer:
            raise aspecd.exceptions.MissingImporterError(
                "No importer provided"
            )
        importer.import_into(self)
        self._origdata = copy.deepcopy(self.data)

    def export_to(self, exporter=None):
        """Export data and metadata.

        This requires initialising an :obj:`aspecd.io.DatasetImporter` object
        first that is provided as an argument for this method.

        .. note::
            The same operation can be performed by calling the
            :meth:`export_from` method of an :obj:`aspecd.io.Exporter`
            object taking an :obj:`aspecd.dataset.Dataset` object as argument.

            However, as usually the dataset is already at hand,
            first creating an instance of a respective exporter
            and then calling :meth:`export_to` of the dataset is the
            preferred way.

        Parameters
        ----------
        exporter : :class:`aspecd.io.DatasetExporter`
            Exporter writing data and metadata to specific output format

        """
        if not exporter:
            raise aspecd.exceptions.MissingExporterError(
                "No exporter provided"
            )
        exporter.export_from(self)
        exporter_record = exporter.create_history_record()
        self._append_task(kind="export", task=exporter_record)

    def add_reference(self, dataset=None):
        """
        Add a reference to another dataset to the list of references.

        A reference is always an object of type
        :class:`aspecd.dataset.DatasetReference` that will be automatically
        created from the dataset provided.

        Parameters
        ----------
        dataset : :class:`aspecd.dataset.Dataset`
            dataset a reference for should be added to the list of references

        Raises
        ------
        aspecd.exceptions.MissingDatasetError
            Raised if no dataset was provided

        """
        if not dataset:
            raise aspecd.exceptions.MissingDatasetError
        dataset_reference = aspecd.dataset.DatasetReference()
        dataset_reference.from_dataset(dataset=dataset)
        self.references.append(dataset_reference)

    def remove_reference(self, dataset_id=None):
        """
        Remove a reference to another dataset from the list of references.

        A reference is always an object of type
        :class:`aspecd.dataset.DatasetReference` that was automatically
        created from the respective dataset when adding the reference.

        Parameters
        ----------
        dataset_id : :class:`str`
            ID of the dataset the reference should be removed for

        Raises
        ------
        aspecd.exceptions.MissingDatasetError
            Raised if no dataset ID was provided

        """
        if not dataset_id:
            raise aspecd.exceptions.MissingDatasetError
        for index, reference in enumerate(self.references):
            if dataset_id == reference.id:
                del self.references[index]
                break

    def from_dict(self, dict_=None):  # noqa: MC0001
        """
        Set properties from dictionary.

        Only parameters in the dictionary that are valid properties of the
        class are set accordingly.

        .. note::
            In conjunction with the :meth:`aspecd.dataset.to_dict` method,
            this method allows to serialise and deserialise dataset objects,
            *i.e.* all kinds of storage to the persistence layer.


        Parameters
        ----------
        dict_ : :class:`dict`
            Dictionary containing properties to set

        """
        for key in dict_:
            if hasattr(self, key):
                attribute = getattr(self, key)
                if key == "history":
                    for element in dict_[key]:
                        record = aspecd.history.ProcessingHistoryRecord()
                        record.from_dict(element)
                        self.history.append(record)
                elif key == "analyses":
                    for element in dict_[key]:
                        record = aspecd.history.AnalysisHistoryRecord()
                        record.from_dict(element)
                        self.analyses.append(record)
                elif key == "annotations":
                    for element in dict_[key]:
                        record = aspecd.history.AnnotationHistoryRecord()
                        record.from_dict(element)
                        self.annotations.append(record)
                elif key == "representations":
                    for element in dict_[key]:
                        record = aspecd.history.PlotHistoryRecord()
                        record.from_dict(element)
                        self.representations.append(record)
                elif key == "references":
                    for element in dict_[key]:
                        record = DatasetReference()
                        record.from_dict(element)
                        self.references.append(record)
                elif key == "tasks":
                    for element in dict_[key]:
                        if element["kind"] == "representation":
                            record_class_name = (
                                "aspecd.history.PlotHistoryRecord"
                            )
                        elif element["kind"] == "annotation":
                            record_class_name = (
                                "aspecd.history.AnnotationHistoryRecord"
                            )
                        else:
                            record_class_name = (
                                "aspecd.history."
                                + element["kind"].capitalize()
                                + "HistoryRecord"
                            )
                        record = aspecd.utils.object_from_class_name(
                            record_class_name
                        )
                        # noinspection PyUnresolvedReferences
                        record.from_dict(element["task"])
                        self.tasks.append(
                            {"kind": element["kind"], "task": record}
                        )
                elif hasattr(attribute, "from_dict"):
                    attribute.from_dict(dict_[key])
                else:
                    setattr(self, key, dict_[key])


class ExperimentalDataset(Dataset):
    """Base class for experimental datasets.

    The dataset is one of the core elements of the ASpecD framework, basically
    containing both, (numeric) data and corresponding metadata, aka information
    available about the data.

    The public attributes of a dataset can be converted to a dict via
    :meth:`aspecd.utils.ToDictMixin.to_dict()`.

    Attributes
    ----------
    metadata : :obj:`aspecd.metadata.ExperimentalDatasetMetadata`
        hierarchical key-value store of metadata

    """

    def __init__(self):
        super().__init__()
        self.metadata = aspecd.metadata.ExperimentalDatasetMetadata()


class CalculatedDataset(Dataset):
    """Base class for datasets containing calculated data.

    The dataset is one of the core elements of the ASpecD framework, basically
    containing both, (numeric) data and corresponding metadata, aka information
    available about the data.

    The public attributes of a dataset can be converted to a dict via
    :meth:`aspecd.utils.ToDictMixin.to_dict()`.

    Attributes
    ----------
    metadata : :obj:`aspecd.metadata.CalculatedDatasetMetadata`
        hierarchical key-value store of metadata

    """

    def __init__(self):
        super().__init__()
        self.data.calculated = True
        self._origdata.calculated = True
        self.metadata = aspecd.metadata.CalculatedDatasetMetadata()


class DatasetReference(aspecd.utils.ToDictMixin):
    """
    Reference to a given dataset.

    Often, one dataset needs to reference other datasets. A typical example
    would be a simulation stored in a dataset of class
    :class:`aspecd.dataset.CalculatedDataset` that needs to reference the
    corresponding experimental data, stored in a dataset of class
    :class:`aspecd.dataset.ExperimentalDataset`. Vice versa,
    the experimental dataset might want to store a reference to one (or
    more) simulations.

    As the dataset ID is not sufficient, both, the ID as well as the
    history of the dataset at the time the reference has been created gets
    stored in the reference and restored upon creating a (new) dataset.
    Hence, at least the data of the dataset returned should be identical to
    the data of the original dataset the reference has been created for.

    Attributes
    ----------
    type : :class:`str`
        type of dataset

        Will be inferred directly from dataset when creating a reference
        from a given dataset and is used to return a dataset of same type.
    id : :class:`str`
        (unique) id of the dataset, i.e. path, LOI, or else
    history : :class:`list`
        history of processing steps performed on the dataset to be referenced

    Raises
    ------
    aspecd.exceptions.MissingDatasetError
        Raised if no dataset was provided when calling :meth:`from_dataset`

    """

    def __init__(self):
        super().__init__()
        self.type = ""
        self.id = ""  # pylint: disable=invalid-name
        self.history = []

    def from_dataset(self, dataset=None):
        """
        Create dataset reference from dataset.

        Parameters
        ----------
        dataset : :class:`aspecd.dataset.Dataset`
            Dataset the reference should be created for

        Raises
        ------
        aspecd.exceptions.MissingDatasetError
            Raised if no dataset was provided

        """
        if not dataset:
            raise aspecd.exceptions.MissingDatasetError
        self.type = aspecd.utils.full_class_name(dataset)
        self.id = dataset.id
        self.history = copy.deepcopy(dataset.history)

    def to_dataset(self):
        """
        Create (new) dataset from reference

        The history stored will be applied to the newly created dataset,
        hence the dataset should be in the same state with respect to
        processing steps as the original dataset was upon creating the
        reference.

        Returns
        -------
        dataset : :class:`aspecd.dataset.Dataset`
            Dataset with identical data to the one the reference has been
            created for

        """
        if not self.type:
            raise aspecd.exceptions.MissingDatasetError
        dataset = aspecd.utils.object_from_class_name(self.type)
        dataset.id = self.id
        for history_record in self.history:
            history_record.replay(dataset)
        return dataset

    def from_dict(self, dict_=None):
        """
        Set properties from dictionary.

        Only parameters in the dictionary that are valid properties of the
        class are set accordingly.

        Parameters
        ----------
        dict_ : :class:`dict`
            Dictionary containing properties to set

        """
        for key, value in dict_.items():
            if hasattr(self, key):
                setattr(self, key, value)


class DatasetFactory:
    """
    Factory for creating dataset objects based on the source provided.

    Particularly in case of recipe-driven data analysis (c.f. :mod:`tasks`),
    there is a need to automatically retrieve datasets using nothing more
    than a source string that can be, e.g., a path or LOI.

    Packages derived from ASpecD should implement a :class:`DatasetFactory`
    inheriting from :class:`aspecd.dataset.DatasetFactory` and overriding
    the protected method :meth:`_create_dataset`. The only task of this
    protected method is to provide the correct dataset object, in most
    cases an instance of a class inheriting from
    :class:`aspecd.dataset.ExperimentalDataset`.

    Attributes
    ----------
    importer_factory : :class:`aspecd.io.DatasetImporterFactory`
        ImporterFactory instance used for importing datasets

    Raises
    ------
    aspecd.exceptions.MissingSourceError
        Raised if no source is provided
    aspecd.exceptions.MissingImporterFactoryError
        Raised if no ImporterFactory is available

    """

    def __init__(self):
        self.importer_factory = None

    def get_dataset(self, source="", importer="", parameters=None):
        """
        Return dataset object for dataset specified by its source.

        The import of data into the dataset is handled using an instance of
        :class:`aspecd.io.DatasetImporterFactory`.

        The actual code for deciding which type of dataset to return in what
        case should be implemented in the non-public method
        :meth:`_create_dataset` in any package based on the ASpecD framework.

        Parameters
        ----------
        source : :class:`str`
            string describing the source of the dataset

            May be a filename or path, a URL/URI, a LOI, or similar

        importer : :class:`str`
            Name of the importer to use for importing the dataset

            Default: ''

            .. versionadded:: 0.2

        parameters : :class:`dict`
            Additional parameters for controlling the import

            Default: None

            .. versionadded:: 0.2

        Returns
        -------
        dataset : :class:`aspecd.dataset.Dataset`
            Dataset object of appropriate class

        Raises
        ------
        aspecd.exceptions.MissingSourceError
            Raised if no source is provided
        aspecd.exceptions.MissingImporterFactoryError
            Raised if no ImporterFactory is available

        """
        if not source:
            raise aspecd.exceptions.MissingSourceError(
                "A source is required to return a dataset"
            )
        if not self.importer_factory:
            raise aspecd.exceptions.MissingImporterFactoryError(
                "An ImporterFactory is required to return a dataset"
            )
        dataset_ = self._create_dataset(source=source)
        importer = self.importer_factory.get_importer(
            source=source, importer=importer, parameters=parameters
        )
        dataset_.import_from(importer)
        return dataset_

    # noinspection PyUnusedLocal
    # pylint: disable=unused-argument
    @staticmethod
    def _create_dataset(source=""):
        """
        Non-public method creating the actual (empty) dataset object.

        Classes inheriting from :class:`aspecd.dataset.DatasetFactory`
        should return an instance of the appropriate class that should in
        all cases inherit from :class:`aspecd.dataset.Dataset` or one of
        its subclasses.

        The parameter :param:`source` may be used to distinguish which type of
        dataset should be returned.

        Parameters
        ----------
        source : :class:`str`
            string describing the source of the dataset

            May be a filename or path, a URL/URI, a LOI, or similar

        Returns
        -------
        dataset : :class:`aspecd.dataset.Dataset`
            Dataset object of appropriate type

        """
        return ExperimentalDataset()


class Data(aspecd.utils.ToDictMixin):
    """
    Unit containing both, numeric data and corresponding axes.

    The data class ensures consistency in terms of dimensions between
    numerical data and axes.

    Parameters
    ----------
    data : :class:`numpy.ndarray`
        Numerical data
    axes : :class:`list`
        List of objects of type :class:`aspecd.dataset.Axis`

        The number of axes needs to be consistent with the dimensions of data.

        Axes will be set automatically when setting data. Hence,
        the easiest is to first set data and only afterwards set axis values.
    calculated : :class:`bool`
        Indicator for the origin of the numerical data (calculation or
        experiment).

    Attributes
    ----------
    calculated : :class:`bool`
        Indicate whether numeric data are calculated rather than
        experimentally recorded

    Raises
    ------
    aspecd.exceptions.AxesCountError
        Raised if number of axes is inconsistent with data dimensions
    aspecd.exceptions.AxesValuesInconsistentWithDataError
        Raised if axes values are inconsistent with data

    """

    def __init__(self, data=np.zeros(0), axes=None, calculated=False):
        super().__init__()
        self._data = data
        self._axes = []
        if axes is None:
            self._create_axes()
        else:
            self.axes = axes
        self.calculated = calculated
        self._include_in_to_dict = ["data", "axes"]

    @property
    def data(self):
        """Get or set (numeric) data.

        .. note::
            If you set data that have different dimensions to the data
            previously stored in the dataset, the axes values will be
            set to an array with indices corresponding to the size of the
            respective data dimension. You will most probably assign proper
            axis values afterwards. On the other hand, all other
            information stored in the axis object will be retained, namely
            quantity, unit, and label.

        """
        return self._data

    @data.setter
    def data(self, data):
        old_shape = self._data.shape
        self._data = data
        if old_shape != data.shape:
            if self.axes[0].values.size == 0:
                self._create_axes()
            self._update_axes()

    @property
    def axes(self):
        """Get or set axes.

        If you set axes, they will be checked for consistency with the data.
        Therefore, first set the data and only afterwards the axes,
        with values corresponding to the dimensions of the data.

        Raises
        ------
        aspecd.exceptions.AxesCountError
            Raised if number of axes is inconsistent with data dimensions
        aspecd.exceptions.AxesValuesInconsistentWithDataError
            Raised if axes values are inconsistent with data dimensions

        """
        return self._axes

    @axes.setter
    def axes(self, axes):
        self._axes = axes
        self._check_axes()

    def _create_axes(self):
        self._axes = []
        missing_axes = self._data.ndim + 1
        # pylint: disable=unused-variable
        # pylint: disable=invalid-name
        for ax in range(missing_axes):
            self._axes.append(Axis())

    def _update_axes(self):
        """
        Update axes according to data

        .. note::
            It turned out to be a bad idea to automatically *remove* an
            axis, as you usually do not know which axis to remove. Hence,
            if some task reduces the dimensionality of the data, this task
            is responsible to adjust the axes as well (*i.e.*, remove the
            *correct* axis object from the list).
        """
        data_shape = self.data.shape
        if len(self.axes) < self.data.ndim + 1:
            self._axes.append(Axis())
        for index in range(self.data.ndim):
            if len(self.axes[index].values) != data_shape[index]:
                self.axes[index].values = np.arange(
                    data_shape[index], dtype=np.float64
                )

    def _check_axes(self):
        if len(self._axes) > self.data.ndim + 1:
            raise aspecd.exceptions.AxesCountError
        data_shape = self.data.shape
        for index in range(self.data.ndim):
            if len(self.axes[index].values) != data_shape[index]:
                raise aspecd.exceptions.AxesValuesInconsistentWithDataError

    def from_dict(self, dict_=None):
        """
        Set properties from dictionary, e.g., from serialised dataset.

        Only parameters in the dictionary that are valid properties of the
        class are set accordingly.

        The list of axes is handled appropriately.

        Parameters
        ----------
        dict_ : :class:`dict`
            Dictionary containing properties to set

        """
        for key in dict_.keys():
            if key == "data":
                self.data = dict_[key]
            elif key == "axes":
                for number, axis in enumerate(dict_[key]):
                    self.axes[number].from_dict(axis)


class Axis(aspecd.utils.ToDictMixin):
    """Axis for data in a dataset.

    An axis contains always both, numerical values as well as the metadata
    necessary to create axis labels and to make sense of the numerical
    information.

    Attributes
    ----------
    quantity : :class:`str`
        quantity of the numerical data, usually used as first part of an
        automatically generated axis label
    unit : :class:`str`
        unit of the numerical data, usually used as second part of an
        automatically generated axis label
    symbol : :class:`str`
        symbol for the quantity of the numerical data, usually used as first
        part of an automatically generated axis label
    label : :class:`str`
        manual label for the axis, particularly useful in cases where no
        quantity and unit are provided or should be overwritten.


    .. note::
        There are three alternative ways of writing axis labels, one with
        using the quantity name and the unit, one with using the quantity
        symbol and the unit, and one using both, quantity name and symbol,
        usually separated by comma. Quantity and unit shall always be
        separated by a slash. Which way you prefer is a matter of personal
        taste and given context.


    Raises
    ------
    ValueError
        Raised when trying to set axis values to another type than numpy array
    IndexError
        Raised when trying to set axis values to an array with more than one
        dimension.
        Raised if index does not have the same length as values.

    """

    def __init__(self):
        super().__init__()
        self._values = np.zeros(0)
        self._index = []
        self._equidistant = None
        self.quantity = ""
        self.symbol = ""
        self.unit = ""
        self.label = ""
        self._include_in_to_dict = ["values", "index"]

    @property
    def values(self):
        """
        Get or set the numerical axis values.

        Values require to be a one-dimensional numpy array. Trying to set
        values to either a different type that cannot be converted to a
        numpy array or a numpy array with more than one dimension will raise
        a corresponding error.

        Raises
        ------
        ValueError
            Raised if axis values are of wrong type
        IndexError
            Raised if axis values are of wrong dimension, i.e. not a vector

        """
        return self._values

    @values.setter
    def values(self, values):
        if not isinstance(values, type(self._values)):
            values = np.asarray(values)
            if (
                not isinstance(values, type(self._values))
                or values.dtype != self._values.dtype
            ):
                raise ValueError(
                    f"Wrong type: expected {self._values.dtype}, "
                    f"got {values.dtype}"
                )
        if values.ndim > 1:
            raise IndexError("Values need to be one-dimensional")
        self._values = values
        self._set_equidistant_property()
        self._set_index()

    @property
    def index(self):
        """
        Get or set the index corresponding to the axis values.

        The index is a list of data labels for each element in the axis
        values, similar to the index in a :class:`pandas.Series`. However,
        in contrast to pandas, usually the index is a list of empty strings.

        The main reason for introducing the index is to allow for tabular
        representation of (calculated) datasets, *e.g.* as a result of an
        analysis, either including multiple values or spanning multiple
        datasets.

        Raises
        ------
        IndexError
            Raised if index does not have the same length as values


        .. versionadded:: 0.5

        """
        return self._index

    @index.setter
    def index(self, index):
        if len(index) != len(self._values):
            raise IndexError("index and values need to be of same length")
        self._index = index

    def _set_index(self):
        self._index = ["" for _ in self.values]

    @property
    def equidistant(self):
        """Return whether the axes values are equidistant.

        True if the axis values are equidistant, False otherwise. None in
        case of no axis values.

        The property is set automatically if axis values are set and
        therefore read-only.

        While simple plotting of data values against non-uniform axes with
        non-equidistant values is usually straightforward, many processing
        steps rely on equidistant axis values in their simplest possible
        implementation.

        """
        return self._equidistant

    def _set_equidistant_property(self):
        if not self.values.size or not self.values.all():
            return
        differences = self.values[1:] - self.values[0:-1]
        self._equidistant = np.isclose(differences.max(), differences.min())

    def from_dict(self, dict_=None):
        """
        Set properties from dictionary, e.g., from serialised dataset.

        Only parameters in the dictionary that are valid properties of the
        class are set accordingly.

        Parameters
        ----------
        dict_ : :class:`dict`
            Dictionary containing properties to set

        """
        for key in dict_:
            if hasattr(self, key):
                setattr(self, key, dict_[key])


class DeviceData(Data):
    """
    Additional data from devices recorded parallel to the actual data.

    The dataset concept (see :class:`aspecd.dataset.Dataset`) rests on the
    assumption that there is one particular set of data that can be
    regarded as the actual or primary data of the dataset. However,
    in many cases, parallel to these actual data, other data are recorded
    as well, be it readouts from monitors or alike.

    Usually, these additional data will share one axis with the
    primary data of the dataset. However, this is not necessarily the
    case. Furthermore, one dataset may contain an arbitrary number of
    additional device data entries.

    Technically speaking, :class:`aspecd.dataset.DeviceData` are a special
    or extended type of :class:`aspecd.dataset.Data`, *i.e.* a unit
    containing both numerical data and corresponding axes. However,
    this class extends that with metadata specific for the device the
    additional data have been recorded with. Why storing metadata here and
    not in the :attr:`aspecd.dataset.Dataset.metadata` property? The
    latter is more concerned with an overall description of the
    experimental setup in sufficient detail, while the metadata contained
    in this class are more device-specific. Potential contents of the
    metadata here are internal device IDs, addresses for communication,
    and alike. Eventually, the metadata contained herein are those that
    can be relevant mainly for debugging purposes or sanity checks of
    experiments.

    .. admonition:: Example

        A real example for additional data recorded in spectroscopy comes
        from time-resolved EPR (tr-EPR) spectroscopy: Here, you usually
        record 2D data as function of magnetic field and time, *i.e.* a
        full time profile per magnetic field point. As this is a
        non-standard method, often setups are controlled by lab-written
        software and allow for monitoring parameters not usually recorded
        with commercial setups. In this particular case, this can be the
        time stamp and microwave frequency for each individual recorded
        time trace, and the `Python trEPR package
        <https://docs.trepr.de/>`_ not only handles tr-EPR data, but is
        capable of dealing with both additional types of data for analysis
        purposes.


    Attributes
    ----------
    metadata : :class:`aspecd.metadata.Device`
        Metadata of the device used to record the additional data

        .. note::
            For actual devices you will want to create dedicated classes
            inheriting from :class:`aspecd.metadata.Device` and extending
            the available attributes.

    calculated : :class:`bool`
        Indicator for the origin of the numerical data (calculation or
        experiment).

        Default: `False`


    .. versionadded:: 0.9

    """

    def __init__(self):
        super().__init__()
        self.metadata = aspecd.metadata.Device()
        self.calculated = False



import os
import textwrap
import unittest

import numpy as np

from aspecd import table, dataset, exceptions


class TestTable(unittest.TestCase):
    def setUp(self):
        self.table = table.Table()
        self.dataset = dataset.CalculatedDataset()
        self.dataset.data.data = np.random.random(5)
        self.filename = "foo"

    def tearDown(self):
        if os.path.exists(self.filename):
            os.remove(self.filename)

    def test_instantiate_class(self):
        pass

    def test_has_to_dict_method(self):
        self.assertTrue(hasattr(self.table, "to_dict"))
        self.assertTrue(callable(self.table.to_dict))

    def test_to_dict_does_not_contain_certain_keys(self):
        for key in ["name", "dataset", "table"]:
            with self.subTest(key=key):
                self.assertNotIn(key, self.table.to_dict())

    def test_has_tabulate_method(self):
        self.assertTrue(hasattr(self.table, "tabulate"))
        self.assertTrue(callable(self.table.tabulate))

    def test_tabulate_without_dataset_raises(self):
        self.table.dataset = None
        with self.assertRaises(exceptions.MissingDatasetError):
            self.table.tabulate()

    def test_tabulate_with_dataset_with_empty_data_warns(self):
        self.table.dataset = dataset.Dataset()
        with self.assertLogs(__package__, level="WARNING") as cm:
            self.table.tabulate()
        self.assertIn(
            "Dataset contains no data, hence nothing to tabulate.",
            cm.output[0],
        )

    def test_tabulate_with_dataset_parameter_sets_dataset(self):
        self.table.dataset = None
        self.table.tabulate(self.dataset)
        self.assertEqual(self.dataset, self.table.dataset)

    def test_tabulate_with_dataset_adds_representation_to_dataset(self):
        self.table.dataset = None
        self.table.tabulate(self.dataset)
        self.assertTrue(self.dataset.representations)

    def test_tabulate_with_3D_dataset_raises(self):
        self.dataset.data.data = np.random.random([5, 5, 5])
        self.table.dataset = self.dataset
        with self.assertRaisesRegex(
            exceptions.NotApplicableToDatasetError,
            "Tables work only with 1D and 2D data",
        ):
            self.table.tabulate()

    def test_tabulate_sets_table(self):
        self.table.dataset = self.dataset
        self.table.tabulate()
        self.assertTrue(self.table.table)

    def test_table_and_dataset_rows_are_consistent(self):
        self.table.dataset = self.dataset
        self.table.tabulate()
        self.assertEqual(
            self.dataset.data.data.shape[0], len(self.table.table.split("\n"))
        )

    def test_multiple_tabulate_calls_reset_table(self):
        self.table.dataset = self.dataset
        self.table.tabulate()
        self.table.tabulate()
        self.assertEqual(
            self.dataset.data.data.shape[0], len(self.table.table.split("\n"))
        )

    def test_rows_have_equal_width(self):
        self.dataset.data.data = np.asarray([1.0, 3.14, 4.1234])
        self.table.dataset = self.dataset
        self.table.tabulate()
        rows = self.table.table.split("\n")
        width = max([len(x) for x in rows])
        for row in rows:
            self.assertEqual(width, len(row))

    def test_tabulate_with_multiple_rows(self):
        self.dataset.data.data = np.ones(3) * np.pi
        self.table.dataset = self.dataset
        self.table.tabulate()
        column = "{}\n{}\n{}".format(*self.dataset.data.data)
        self.assertEqual(column, self.table.table)

    def test_tabulate_with_multiple_columns(self):
        self.dataset.data.data = np.random.random([1, 3])
        self.table.dataset = self.dataset
        self.table.tabulate()
        row = "{} {} {}".format(*self.dataset.data.data[0, :])
        self.assertEqual(row, self.table.table)

    def test_tabulate_with_column_format(self):
        self.dataset.data.data = np.random.random(3)
        self.table.dataset = self.dataset
        format_ = "5.3f"
        self.table.column_format = [format_]
        self.table.tabulate()
        column = "{:{format}}\n{:{format}}\n{:{format}}".format(
            *self.dataset.data.data, format=format_
        )
        self.assertEqual(column, self.table.table)

    def test_tabulate_with_column_format_and_multiple_columns(self):
        self.dataset.data.data = np.random.random([1, 3])
        self.table.dataset = self.dataset
        format_ = ["5.3f", "6.4f", "7.5f"]
        self.table.column_format = format_
        self.table.tabulate()
        column = "{:{format0}} {:{format1}} {:{format2}}".format(
            *self.dataset.data.data[0, :],
            format0=format_[0],
            format1=format_[1],
            format2=format_[2],
        )
        self.assertEqual(column, self.table.table)

    def test_tabulate_with_one_column_format_and_multiple_columns(self):
        self.dataset.data.data = np.random.random([1, 3])
        self.table.dataset = self.dataset
        format_ = "5.3f"
        self.table.column_format = [format_]
        self.table.tabulate()
        column = "{:{format}} {:{format}} {:{format}}".format(
            *self.dataset.data.data[0, :], format=format_
        )
        self.assertEqual(column, self.table.table)

    def test_index_in_first_axis_adds_index_to_rows(self):
        self.dataset.data.data = np.random.random(3)
        index = ["foo", "bar", "foobar"]
        self.dataset.data.axes[0].index = index
        self.table.dataset = self.dataset
        self.table.tabulate()
        rows = self.table.table.split("\n")
        for idx, row in enumerate(rows):
            self.assertTrue(row.startswith(index[idx]))

    def test_index_in_second_axis_adds_column_headers(self):
        self.dataset.data.data = np.random.random([1, 3])
        index = ["foo", "bar", "foobar"]
        self.dataset.data.axes[1].index = index
        self.table.dataset = self.dataset
        self.table.tabulate()
        column_headers = self.table.table.split("\n")[0].split()
        for idx, header in enumerate(column_headers):
            self.assertEqual(index[idx], header)

    def test_long_column_headers_expand_data_column_width(self):
        self.dataset.data.data = np.random.random([1, 3])
        index = ["foo foo foo foo foo", "bar", "foobar"]
        self.table.column_format = ["5.3f"]
        self.dataset.data.axes[1].index = index
        self.table.dataset = self.dataset
        self.table.tabulate()
        data_rows = self.table.table.split("\n")
        self.assertEqual(len(data_rows[0]), len(data_rows[1]))

    def test_column_headers_with_row_index(self):
        self.dataset.data.data = np.ones([3, 3])
        index = ["foo", "bar", "foobar"]
        self.dataset.data.axes[0].index = index
        self.dataset.data.axes[1].index = index
        self.table.dataset = self.dataset
        self.table.tabulate()
        column_headers = self.table.table.split("\n")[0]
        self.assertTrue(column_headers.startswith("   "))

    @unittest.skip
    def test_column_separator_from_format(self):
        self.dataset.data.data = np.random.random([1, 3])
        self.table.dataset = self.dataset
        format_ = table.Format()
        format_.column_separator = " & "
        self.table.format = format_
        self.table.tabulate()
        row = "{}{sep}{}{sep}{}".format(
            *self.dataset.data.data[0, :], sep=format_.column_separator
        )
        self.assertEqual(row, self.table.table)

    @unittest.skip
    def test_column_prefix_from_format(self):
        self.dataset.data.data = np.random.random([1, 3])
        self.table.dataset = self.dataset
        format_ = table.Format()
        format_.column_prefix = "| "
        format_.column_separator = " | "
        self.table.format = format_
        self.table.tabulate()
        row = "{prefix}{}{sep}{}{sep}{}".format(
            *self.dataset.data.data[0, :],
            sep=format_.column_separator,
            prefix=format_.column_prefix,
        )
        self.assertEqual(row, self.table.table)

    @unittest.skip
    def test_column_postfix_from_format(self):
        self.dataset.data.data = np.random.random([1, 3])
        self.table.dataset = self.dataset
        format_ = table.Format()
        format_.column_postfix = "| "
        format_.column_separator = " | "
        self.table.format = format_
        self.table.tabulate()
        row = "{}{sep}{}{sep}{}{postfix}".format(
            *self.dataset.data.data[0, :],
            sep=format_.column_separator,
            postfix=format_.column_postfix,
        )
        self.assertEqual(row, self.table.table)

    @unittest.skip
    def test_header_separator_from_format(self):
        self.dataset.data.data = np.ones([1, 3])
        index = ["foo", "bar", "foobar"]
        self.dataset.data.axes[1].index = index
        self.table.dataset = self.dataset
        format_ = table.Format()
        format_.header_separator = " ^ "
        self.table._format = format_
        self.table.tabulate()
        row = "{}{sep}{}{sep}{}".format(*index, sep=format_.header_separator)
        self.assertEqual(row, self.table.table.split("\n")[0])

    @unittest.skip
    def test_header_prefix_from_format(self):
        self.dataset.data.data = np.ones([1, 3])
        index = ["foo", "bar", "foobar"]
        self.dataset.data.axes[1].index = index
        self.table.dataset = self.dataset
        format_ = table.Format()
        format_.header_prefix = "^ "
        format_.header_separator = " ^ "
        self.table.format = format_
        self.table.tabulate()
        row = "{prefix}{}{sep}{}{sep}{}".format(
            *index, sep=format_.header_separator, prefix=format_.header_prefix
        )
        self.assertEqual(row, self.table.table.split("\n")[0])

    @unittest.skip
    def test_header_postfix_from_format(self):
        self.dataset.data.data = np.ones([1, 3])
        index = ["foo", "bar", "foobar"]
        self.dataset.data.axes[1].index = index
        self.table.dataset = self.dataset
        format_ = table.Format()
        format_.header_postfix = "^ "
        format_.header_separator = " ^ "
        self.table.format = format_
        self.table.tabulate()
        row = "{}{sep}{}{sep}{}{postfix}".format(
            *index,
            sep=format_.header_separator,
            postfix=format_.header_postfix,
        )
        self.assertEqual(row, self.table.table.split("\n")[0])

    def test_top_rule(self):
        self.dataset.data.data = np.ones([1, 3])
        self.table.dataset = self.dataset
        format_ = table.TextFormat()
        self.table.format = "text"
        self.table.tabulate()
        row = format_.top_rule(column_widths=[3, 3, 3])
        self.assertEqual(row, self.table.table.split("\n")[0])

    def test_middle_rule(self):
        self.dataset.data.data = np.ones([3, 3])
        index = ["foo", "bar", "baz"]
        self.dataset.data.axes[1].index = index
        self.table.dataset = self.dataset
        format_ = table.TextFormat()
        self.table.format = "text"
        self.table.tabulate()
        row = format_.middle_rule(column_widths=[3, 3, 3])
        self.assertEqual(row, self.table.table.split("\n")[2])

    def test_bottom_rule(self):
        self.dataset.data.data = np.ones([1, 3])
        self.table.dataset = self.dataset
        format_ = table.TextFormat()
        self.table.format = "text"
        self.table.tabulate()
        row = format_.bottom_rule(column_widths=[3, 3, 3])
        self.assertEqual(row, self.table.table.split("\n")[-1])

    def test_padding(self):
        self.dataset.data.data = np.ones([1, 3])
        self.table.dataset = self.dataset
        format_ = table.TextFormat()
        format_.padding = 1
        self.table.format = "text"
        self.table.tabulate()
        row = "{prefix} {} {sep} {} {sep} {} {postfix}".format(
            *self.dataset.data.data[0, :],
            prefix=format_.column_prefix,
            sep=format_.column_separator,
            postfix=format_.column_postfix,
        )
        self.assertEqual(row, self.table.table.split("\n")[1])

    def test_padding_with_header(self):
        self.dataset.data.data = np.ones([1, 3])
        index = ["foo", "bar", "baz"]
        self.dataset.data.axes[1].index = index
        self.table.dataset = self.dataset
        format_ = table.TextFormat()
        format_.padding = 1
        self.table.format = "text"
        self.table.tabulate()
        row = "{prefix} {} {sep} {} {sep} {} {postfix}".format(
            *index,
            prefix=format_.header_prefix,
            sep=format_.header_separator,
            postfix=format_.header_postfix,
        )
        self.assertEqual(row, self.table.table.split("\n")[1])

    def test_dokuwiki_format(self):
        self.dataset.data.data = np.ones([1, 3])
        self.table.dataset = self.dataset
        format_ = table.DokuwikiFormat()
        self.table.format = "dokuwiki"
        self.table.tabulate()
        row = "{prefix} {} {sep} {} {sep} {} {postfix}".format(
            *self.dataset.data.data[0, :],
            prefix=format_.column_prefix,
            sep=format_.column_separator,
            postfix=format_.column_postfix,
        )
        self.assertEqual(row, self.table.table.split("\n")[0])

    def test_dokuwiki_format_header(self):
        self.dataset.data.data = np.ones([1, 3])
        index = ["foo", "bar", "baz"]
        self.dataset.data.axes[1].index = index
        self.table.dataset = self.dataset
        format_ = table.DokuwikiFormat()
        self.table.format = "dokuwiki"
        self.table.tabulate()
        row = "{prefix} {} {sep} {} {sep} {} {postfix}".format(
            *index,
            prefix=format_.header_prefix,
            sep=format_.header_separator,
            postfix=format_.header_postfix,
        )
        self.assertEqual(row, self.table.table.split("\n")[0])

    def test_dokuwiki_format_row_index(self):
        self.dataset.data.data = np.ones(3)
        index = ["foo", "bar", "baz"]
        self.dataset.data.axes[0].index = index
        self.table.dataset = self.dataset
        format_ = table.DokuwikiFormat()
        self.table.format = "dokuwiki"
        self.table.tabulate()
        row = "{prefix} {} {sep} {} {postfix}".format(
            *[index[0], self.dataset.data.data[0]],
            prefix=format_.header_prefix,
            sep=format_.column_separator,
            postfix=format_.column_postfix,
        )
        self.assertEqual(row, self.table.table.split("\n")[0])

    def test_dokuwiki_format_header_and_row_index(self):
        self.dataset.data.data = np.ones([3, 3])
        index = ["foo", "bar", "baz"]
        self.dataset.data.axes[0].index = index
        self.dataset.data.axes[1].index = index
        self.table.dataset = self.dataset
        format_ = table.DokuwikiFormat()
        self.table.format = "dokuwiki"
        self.table.tabulate()
        row = "{prefix} {} {sep} {} {sep} {} {sep} {} {postfix}".format(
            *["   ", *index],
            prefix=format_.column_prefix,
            sep=format_.header_separator,
            postfix=format_.header_postfix,
        )
        self.assertEqual(row, self.table.table.split("\n")[0])

    def test_latex_format_opening(self):
        self.dataset.data.data = np.ones([1, 3])
        self.table.dataset = self.dataset
        format_ = table.LatexFormat()
        self.table.format = "latex"
        self.table.tabulate()
        row = r"\begin{tabular}{lll}"
        self.assertEqual(row, self.table.table.split("\n")[0])

    def test_latex_format_closing(self):
        self.dataset.data.data = np.ones([1, 3])
        self.table.dataset = self.dataset
        format_ = table.LatexFormat()
        self.table.format = "latex"
        self.table.tabulate()
        row = r"\end{tabular}"
        self.assertEqual(row, self.table.table.split("\n")[-1])

    def test_latex_format_of_numerical_data(self):
        self.dataset.data.data = np.ones([1, 3])
        self.table.dataset = self.dataset
        format_ = table.LatexFormat()
        self.table.format = "latex"
        self.table.tabulate()
        row = r"{}{sep}{}{sep}{} \\".format(
            *self.dataset.data.data[0, :],
            sep=format_.column_separator,
        )
        self.assertEqual(row, self.table.table.split("\n")[2])

    def test_latex_format_with_header(self):
        self.dataset.data.data = np.ones([1, 3])
        index = ["foo", "bar", "baz"]
        self.dataset.data.axes[1].index = index
        self.table.dataset = self.dataset
        format_ = table.LatexFormat()
        self.table.format = "latex"
        self.table.tabulate()
        self.assertEqual(r"\midrule", self.table.table.split("\n")[3])

    def test_with_caption(self):
        caption = table.Caption()
        caption.title = "Lorem ipsum dolor sit amet"
        self.dataset.data.data = np.ones([3, 3])
        index = ["foo", "bar", "baz"]
        self.dataset.data.axes[0].index = index
        self.dataset.data.axes[1].index = index
        self.table.dataset = self.dataset
        self.table.caption = caption
        self.table.tabulate()
        self.assertEqual(caption.title, self.table.table.split("\n")[0])
        self.assertEqual("", self.table.table.split("\n")[1])

    def test_save(self):
        self.table.dataset = self.dataset
        self.table.tabulate()
        self.table.filename = self.filename
        self.table.save()
        self.assertTrue(os.path.exists(self.filename))

    def test_save_writes_content(self):
        self.table.dataset = self.dataset
        self.table.tabulate()
        self.table.filename = self.filename
        self.table.save()
        with open(self.filename, "r") as file:
            file_content = file.read()
        self.assertEqual(self.table.table, file_content)


class TestFormat(unittest.TestCase):
    def setUp(self):
        self.format = table.Format()

    def test_instantiate_class(self):
        pass

    def test_has_properties(self):
        properties = [
            "column_separator",
            "column_prefix",
            "column_postfix",
            "header_separator",
            "header_prefix",
            "header_postfix",
        ]
        for prop in properties:
            with self.subTest(property=prop):
                self.assertTrue(hasattr(self.format, prop))

    def test_has_top_rule_method(self):
        self.assertTrue(hasattr(self.format, "top_rule"))
        self.assertTrue(callable(self.format.top_rule))

    def test_top_rule_returns_string(self):
        self.assertIsInstance(self.format.top_rule(), str)

    def test_top_rule_with_column_widths(self):
        self.format.top_rule(column_widths=[3, 3, 3])

    def test_has_middle_rule_method(self):
        self.assertTrue(hasattr(self.format, "middle_rule"))
        self.assertTrue(callable(self.format.middle_rule))

    def test_midrule_returns_string(self):
        self.assertIsInstance(self.format.middle_rule(), str)

    def test_middle_rule_with_column_widths(self):
        self.format.middle_rule(column_widths=[3, 3, 3])

    def test_has_bottom_rule_method(self):
        self.assertTrue(hasattr(self.format, "bottom_rule"))
        self.assertTrue(callable(self.format.bottom_rule))

    def test_bottom_rule_returns_string(self):
        self.assertIsInstance(self.format.bottom_rule(), str)

    def test_bottom_rule_with_column_widths(self):
        self.format.bottom_rule(column_widths=[3, 3, 3])

    def test_has_opening_method(self):
        self.assertTrue(hasattr(self.format, "opening"))
        self.assertTrue(callable(self.format.opening))

    def test_opening_returns_string(self):
        self.assertIsInstance(self.format.opening(), str)

    def test_opening_with_columns(self):
        self.format.opening(columns=5)

    def test_opening_with_columns_and_caption(self):
        self.format.opening(columns=5, caption=table.Caption())

    def test_opening_with_caption_returns_caption(self):
        caption = table.Caption()
        caption.title = "Lorem ipsum"
        expected_opening = "\n".join(textwrap.wrap(caption.title)) + "\n"
        opening = self.format.opening(caption=caption)
        self.assertEqual(expected_opening, opening)

    def test_opening_with_caption_with_title_and_text(self):
        caption = table.Caption()
        caption.title = "Lorem ipsum"
        caption.text = (
            "dolor sit amet, consectetur adipiscing elit. "
            "Curabitur commodo tempor vulputate. Sed placerat "
            "ultricies ex. Donec iaculis ipsum non felis mattis, "
            "quis mollis urna pharetra"
        )
        expected_opening = (
            "\n".join(textwrap.wrap(" ".join([caption.title, caption.text])))
            + "\n"
        )
        opening = self.format.opening(caption=caption)
        self.assertEqual(expected_opening, opening)

    def test_has_closing_method(self):
        self.assertTrue(hasattr(self.format, "closing"))
        self.assertTrue(callable(self.format.closing))

    def test_closing_with_caption(self):
        self.format.closing(caption=table.Caption())

    def test_begin_environment_returns_string(self):
        self.assertIsInstance(self.format.closing(), str)


class TestTextFormat(unittest.TestCase):
    def setUp(self):
        self.format = table.TextFormat()

    def test_instantiate_class(self):
        pass

    def test_top_rule_returns_rule(self):
        self.format.padding = 0
        rule = "+---+---+---+"
        self.assertEqual(rule, self.format.top_rule([3, 3, 3]))

    def test_middle_rule_returns_rule(self):
        self.format.padding = 0
        rule = "+---+---+---+"
        self.assertEqual(rule, self.format.middle_rule([3, 3, 3]))

    def test_bottom_rule_returns_rule(self):
        self.format.padding = 0
        rule = "+---+---+---+"
        self.assertEqual(rule, self.format.bottom_rule([3, 3, 3]))

    def test_top_rule_with_padding(self):
        self.format.padding = 2
        rule = "+-------+-------+-------+"
        self.assertEqual(rule, self.format.top_rule([3, 3, 3]))

    def test_default_padding_is_one(self):
        rule = "+-----+-----+-----+"
        self.assertEqual(rule, self.format.top_rule([3, 3, 3]))


class TestRstFormat(unittest.TestCase):
    def setUp(self):
        self.format = table.RstFormat()

    def test_instantiate_class(self):
        pass

    def test_top_rule_returns_rule(self):
        rule = "=== === ==="
        self.assertEqual(rule, self.format.top_rule([3, 3, 3]))

    def test_middle_rule_returns_rule(self):
        rule = "=== === ==="
        self.assertEqual(rule, self.format.middle_rule([3, 3, 3]))

    def test_bottom_rule_returns_rule(self):
        rule = "=== === ==="
        self.assertEqual(rule, self.format.bottom_rule([3, 3, 3]))


class TestDokuwikiFormat(unittest.TestCase):
    def setUp(self):
        self.format = table.DokuwikiFormat()

    def test_instantiate_class(self):
        pass

    def test_top_rule_is_empty(self):
        self.assertEqual("", self.format.top_rule())

    def test_middle_rule_is_empty(self):
        self.assertEqual("", self.format.middle_rule())

    def test_bottom_rule_is_empty(self):
        self.assertEqual("", self.format.bottom_rule())

    def test_opening_without_caption_is_empty(self):
        self.assertEqual("", self.format.opening())

    def test_opening_with_caption(self):
        caption = table.Caption()
        caption.title = "Lorem ipsum"
        expected_opening = "\n".join(
            ["<table>", "<caption>*{}*</caption>".format(caption.title)]
        )
        opening = self.format.opening(caption=caption)
        self.assertEqual(expected_opening, opening)

    def test_opening_with_caption_with_title_and_text(self):
        caption = table.Caption()
        caption.title = "Lorem ipsum"
        caption.text = "dolor sit amet"
        expected_opening = "\n".join(
            [
                "<table>",
                "<caption>*{}* {}</caption>".format(
                    caption.title, caption.text
                ),
            ]
        )
        opening = self.format.opening(caption=caption)
        self.assertEqual(expected_opening, opening)

    def test_closing_without_caption_is_empty(self):
        self.assertEqual("", self.format.closing())

    def test_closing_with_caption(self):
        caption = table.Caption()
        caption.title = "Lorem ipsum"
        closing = self.format.closing(caption=caption)
        self.assertEqual("</table>", closing)


class TestLatexFormat(unittest.TestCase):
    def setUp(self):
        self.format = table.LatexFormat()

    def test_instantiate_class(self):
        pass

    def test_top_rule(self):
        self.assertEqual(r"\toprule", self.format.top_rule())

    def test_middle_rule(self):
        self.assertEqual(r"\midrule", self.format.middle_rule())

    def test_bottom_rule(self):
        self.assertEqual(r"\bottomrule", self.format.bottom_rule())

    def test_opening_with_caption_adds_table_environment_and_caption(self):
        caption = table.Caption()
        caption.title = "Lorem ipsum"
        expected_opening = "\n".join(
            [
                r"\begin{table}",
                r"\caption{\textbf{" + caption.title + r"}}",
                r"\begin{tabular}{l}",
            ]
        )
        opening = self.format.opening(columns=1, caption=caption)
        self.assertEqual(expected_opening, opening)

    def test_opening_with_caption_title_and_text(self):
        caption = table.Caption()
        caption.title = "Lorem ipsum"
        caption.text = "dolor sit amet"
        caption_text = (
            r"\caption{\textbf{" + caption.title + r"} " + caption.text + r"}"
        )
        opening = self.format.opening(columns=1, caption=caption)
        self.assertEqual(caption_text, opening.splitlines()[1])

    def test_closing_with_caption_adds_table_environment(self):
        caption = table.Caption()
        caption.title = "Lorem ipsum"
        expected_closing = "\n".join([r"\end{tabular}", r"\end{table}"])
        closing = self.format.closing(caption=caption)
        self.assertEqual(expected_closing, closing)


class TestCaption(unittest.TestCase):
    def setUp(self):
        self.caption = table.Caption()

    def test_instantiate_class(self):
        pass

    def test_has_to_dict_method(self):
        self.assertTrue(hasattr(self.caption, "to_dict"))
        self.assertTrue(callable(self.caption.to_dict))

    def test_has_from_dict_method(self):
        self.assertTrue(hasattr(self.caption, "from_dict"))
        self.assertTrue(callable(self.caption.from_dict))

    def test_has_title_property(self):
        self.assertTrue(hasattr(self.caption, "title"))

    def test_has_text_property(self):
        self.assertTrue(hasattr(self.caption, "text"))



import aspecd.dataset
import aspecd.utils


class_names = ["Dataset", "ExperimentalDataset", "CalculatedDataset"]

for class_name in class_names:
    yaml = aspecd.utils.Yaml()
    ds = aspecd.utils.object_from_class_name(
        ".".join(["aspecd.dataset", class_name])
    )
    if class_name == "ExperimentalDataset":
        ds.device_data["example"] = aspecd.dataset.DeviceData()
    yaml.dict = ds.to_dict()
    yaml.serialise_numpy_arrays()
    yaml.write_to(".".join([class_name, "yaml"]))
